import tensorflow.keras.optimizers as tf_optimizers
from tensorflow.keras.optimizers.schedules import ExponentialDecay
from ml4ir.base.model.optimizers import cyclic_learning_rate
from ml4ir.base.config.keys import OptimizerKey, LearningRateScheduleKey, CyclicLearningRateType
import tensorflow as tf


class OptimizerDefaultValues(object):
    """Default values for unspecified parameters of the optimizer and learning rate schedule."""

    CONSTANT_LR = 0.01
    EXP_DECAY_STEPS = 100000
    EXP_DECAY_RATE = 0.96
    CYCLIC_INITIAL_LEARNING_RATE = 0.001
    CYCLIC_MAXIMAL_LEARNING_RATE = 0.01
    CYCLIC_STEP_SIZE = 10
    CYCLIC_GAMMA = 1.0


def choose_optimizer(model_config, learning_rate_schedule):
    """
    Define the optimizer used for training the RelevanceModel
    Users have the option to define an ExponentialDecay learning rate schedule

    Parameters
    ----------
    model_config : dict
        model configuration doctionary

    Returns
    -------
    tensorflow optimizer

    Notes
    -----
    References:
    - https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/Optimizer
    - https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/schedules/ExponentialDecay
    - https://arxiv.org/pdf/1506.01186.pdf
    """

    if 'optimizer' not in model_config:
        return tf_optimizers.Adam(learning_rate=learning_rate_schedule, clipvalue=5.0)
    else:
        optimizer_key = model_config['optimizer']['key']
        if 'gradient_clip_value' in model_config['optimizer']:
            config = {'learning_rate': learning_rate_schedule,
                      'clipvalue': model_config['optimizer']['gradient_clip_value']}
        else:
            config = {'learning_rate': learning_rate_schedule}
        return tf.keras.optimizers.get({'class_name': optimizer_key, 'config': config})


def choose_scheduler(model_config):
    """
    Define the optimizer used for training the RelevanceModel
    Users have the option to define an ExponentialDecay learning rate schedule

    Parameters
    ----------
    model_config : dict
        model configuration doctionary

    Returns
    -------
    tensorflow learning rate scheduler

    Notes
    -----
    References:
        https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/Optimizer
        https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/schedules/ExponentialDecay
        https://arxiv.org/pdf/1506.01186.pdf
    """

    if 'lr_schedule' not in model_config:
        # use constant lr schedule
        learning_rate_schedule = ExponentialDecay(
            initial_learning_rate=OptimizerDefaultValues.CONSTANT_LR,
            decay_steps=10000000,
            decay_rate=1.0,
        )

    else:
        lr_schedule = model_config['lr_schedule']
        lr_schedule_key = lr_schedule['key']

        if lr_schedule_key == LearningRateScheduleKey.EXPONENTIAL:
            learning_rate_schedule = ExponentialDecay(
                initial_learning_rate=lr_schedule.get(
                    'learning_rate', OptimizerDefaultValues.CONSTANT_LR),
                decay_steps=lr_schedule.get(
                    'learning_rate_decay_steps', OptimizerDefaultValues.EXP_DECAY_STEPS),
                decay_rate=lr_schedule.get(
                    'learning_rate_decay', OptimizerDefaultValues.EXP_DECAY_RATE),
                staircase=True,
            )

        elif lr_schedule_key == LearningRateScheduleKey.CONSTANT:
            learning_rate_schedule = ExponentialDecay(
                initial_learning_rate=lr_schedule.get(
                    'learning_rate', OptimizerDefaultValues.CONSTANT_LR),
                decay_steps=10000000,
                decay_rate=1.0,
            )
        elif lr_schedule_key == LearningRateScheduleKey.REDUCE_LR_ON_PLATEAU:
            learning_rate_schedule = lr_schedule.get('learning_rate', OptimizerDefaultValues.CONSTANT_LR)

        elif lr_schedule_key == LearningRateScheduleKey.CYCLIC:
            lr_schedule_type = lr_schedule['type']
            if lr_schedule_type == CyclicLearningRateType.TRIANGULAR:
                learning_rate_schedule = cyclic_learning_rate.TriangularCyclicalLearningRate(
                    initial_learning_rate=lr_schedule.get(
                        'initial_learning_rate', OptimizerDefaultValues.CYCLIC_INITIAL_LEARNING_RATE),
                    maximal_learning_rate=lr_schedule.get(
                        'maximal_learning_rate', OptimizerDefaultValues.CYCLIC_MAXIMAL_LEARNING_RATE),
                    step_size=lr_schedule.get(
                        'step_size', OptimizerDefaultValues.CYCLIC_STEP_SIZE),
                )
            elif lr_schedule_type == CyclicLearningRateType.TRIANGULAR2:
                learning_rate_schedule = cyclic_learning_rate.Triangular2CyclicalLearningRate(
                    initial_learning_rate=lr_schedule.get(
                        'initial_learning_rate', OptimizerDefaultValues.CYCLIC_INITIAL_LEARNING_RATE),
                    maximal_learning_rate=lr_schedule.get(
                        'maximal_learning_rate', OptimizerDefaultValues.CYCLIC_MAXIMAL_LEARNING_RATE),
                    step_size=lr_schedule.get(
                        'step_size', OptimizerDefaultValues.CYCLIC_STEP_SIZE),
                )
            elif lr_schedule_type == CyclicLearningRateType.EXPONENTIAL:
                learning_rate_schedule = cyclic_learning_rate.ExponentialCyclicalLearningRate(
                    initial_learning_rate=lr_schedule.get(
                        'initial_learning_rate', OptimizerDefaultValues.CYCLIC_INITIAL_LEARNING_RATE),
                    maximal_learning_rate=lr_schedule.get(
                        'maximal_learning_rate', OptimizerDefaultValues.CYCLIC_MAXIMAL_LEARNING_RATE),
                    step_size=lr_schedule.get(
                        'step_size', OptimizerDefaultValues.CYCLIC_STEP_SIZE),
                    gamma=lr_schedule.get(
                        'gamma', OptimizerDefaultValues.CYCLIC_GAMMA),
                )
            else:
                raise ValueError(
                    "Unsupported cyclic learning rate schedule type key: " + lr_schedule_type)
        else:
            raise ValueError(
                "Unsupported learning rate schedule key: " + lr_schedule_key)

    return learning_rate_schedule


def get_optimizer(model_config) -> tf_optimizers.Optimizer:
    """
    Define the optimizer used for training the RelevanceModel
    Users have the option to define an ExponentialDecay learning rate schedule

    Parameters
    ----------
    model_config_file : str
        Path to model config file

    Returns
    -------
    tensorflow keras optimizer

    Notes
    -----
    References:
        https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/Optimizer
        https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/schedules/ExponentialDecay
        https://arxiv.org/pdf/1506.01186.pdf
    """
    learning_rate_schedule = choose_scheduler(model_config)
    return choose_optimizer(model_config, learning_rate_schedule)



import os
import tensorflow as tf

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.data import tfrecord_reader, tfrecord_writer
from ml4ir.base.io.file_io import FileIO

from typing import List

# Constants
TFRECORD_FILE = "file_0.tfrecord"


def read(
    data_dir: str,
    feature_config: FeatureConfig,
    tfrecord_type: str,
    tfrecord_dir: str,
    file_io: FileIO,
    batch_size: int = 128,
    preprocessing_keys_to_fns: dict = {},
    use_part_files: bool = False,
    max_sequence_size: int = 25,
    parse_tfrecord: bool = True,
    logger=None,
    **kwargs
) -> tf.data.TFRecordDataset:
    """
    Create a TFRecordDataset from directory of CSV files using the FeatureConfig

    Current execution plan:
        1. Load CSVs as pandas dataframes
        2. Convert each query into tf.train.SequenceExample protobufs
        3. Write the protobufs into a .tfrecord file
        4. Load .tfrecord file into a TFRecordDataset and parse the protobufs

    Parameters
    ----------
    data_dir : str
        Path to directory containing csv files to read
    feature_config : FeatureConfig object
        FeatureConfig object that defines the features to be loaded in the dataset
        and the preprocessing functions to be applied to each of them
    tfrecord_dir : str
        Path to directory where the serialized .tfrecord files will be stored
    batch_size : int
        value specifying the size of the data batch
    use_part_files : bool
        load dataset from part files checked using "part-" prefix
    max_sequence_size : int
        value specifying max number of records per query
    logger : Logger object
        logging handler to print and save status messages

    Returns
    -------
    `TFRecordDataset` object
        tensorflow TFRecordDataset loaded from the CSV file
    """
    csv_files: List[str] = file_io.get_files_in_directory(
        data_dir,
        extension="" if use_part_files else ".csv",
        prefix="part-" if use_part_files else "",
    )

    # Create a directory for storing tfrecord files
    file_io.make_directory(tfrecord_dir, clear_dir=True)

    # Write tfrecord files
    tfrecord_writer.write_from_files(
        csv_files=csv_files,
        tfrecord_file=os.path.join(tfrecord_dir, TFRECORD_FILE),
        feature_config=feature_config,
        tfrecord_type=tfrecord_type,
        file_io=file_io,
        logger=logger,
    )

    dataset = tfrecord_reader.read(
        data_dir=tfrecord_dir,
        feature_config=feature_config,
        tfrecord_type=tfrecord_type,
        max_sequence_size=max_sequence_size,
        batch_size=batch_size,
        preprocessing_keys_to_fns=preprocessing_keys_to_fns,
        parse_tfrecord=parse_tfrecord,
        file_io=file_io,
        logger=logger,
    )

    return dataset



import unittest
import warnings
import os
from ml4ir.base.io import logging_utils
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.features.preprocessing import *
from ml4ir.base.model.optimizers import cyclic_learning_rate
from ml4ir.base.config.keys import DataFormatKey, TFRecordTypeKey
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.model.optimizers.optimizer import get_optimizer
from ml4ir.applications.ranking.model.ranking_model import RankingModel
from ml4ir.applications.ranking.config.keys import LossKey
from ml4ir.applications.ranking.config.keys import ScoringTypeKey
from ml4ir.applications.ranking.model.losses import loss_factory
import yaml
from ml4ir.base.features.feature_config import ExampleFeatureConfig, SequenceExampleFeatureConfig
import tensorflow as tf
from tensorflow import keras
import numpy as np
from tensorflow.keras.optimizers.schedules import ExponentialDecay

warnings.filterwarnings("ignore")

INPUT_DIR = "ml4ir/applications/ranking/tests/data/"
MODEL_CONFIG = "ml4ir/applications/ranking/tests/data/configs/model_config_cyclic_lr.yaml"
MODEL_CONFIG_REDUCE_LR_ON_PLATEAU = "ml4ir/applications/ranking/tests/data/configs/model_config_reduce_lr_on_plateau.yaml"
KEEP_ADDITIONAL_INFO = 0
NON_ZERO_FEATURES_ONLY = 0

class LrCallback(keras.callbacks.Callback):
    """Defining a callback to keep track of changing learning rate during training."""

    def __init__(self):
        self.lr_list = []
        self.lr_reduce_on_plateau_list = []

    def on_train_batch_begin(self, batch, logs=None):
        self.lr_list.append(self.model.optimizer._decayed_lr(tf.float32).numpy())

    def on_epoch_begin(self, epoch, logs=None):
        self.lr_reduce_on_plateau_list.append(self.model.optimizer._decayed_lr(tf.float32).numpy())

    def get_lr_list(self):
        return self.lr_list

    def get_lr_reduce_on_plateau_list(self):
        return self.lr_reduce_on_plateau_list

class TestLrSchedules(unittest.TestCase):
    """Testing different learning rate schedules"""

    def setUp(self):
        self.feature_config_yaml_convert_to_clicks = INPUT_DIR + \
            'ranklib/feature_config_convert_to_clicks.yaml'
        self.model_config_file = MODEL_CONFIG

    def compare_lr_values(self, scheduler, expected_values):
        """Expects a learning rate scheduler `scheduler` and
        a list of values. Compares ...."""
        actual_values = [scheduler[i].numpy() for i in range(len(expected_values))]
        return np.all(np.isclose(actual_values, expected_values, rtol=0.001))

    def parse_config(self, tfrecord_type: str, feature_config, io) -> SequenceExampleFeatureConfig:
        if feature_config.endswith(".yaml"):
            feature_config = io.read_yaml(feature_config)
        else:
            feature_config = yaml.safe_load(feature_config)

        return SequenceExampleFeatureConfig(feature_config, None)

    def test_cyclic_tri_lr(self):
        """Test a triangular cyclic learning rate"""
        gold = [0.1,0.28000003,0.45999995,0.64000005,0.81999993,1.0,0.81999993,0.64000005,0.45999995,0.28000003,
                0.1,0.28000003,0.46000007,0.6399999,0.81999993,1.0,0.81999993,0.6399999,0.46000007,0.28000003,
                0.1,0.27999982,0.46000007,0.6399999,0.8200002]
        lrs = cyclic_learning_rate.TriangularCyclicalLearningRate(
            initial_learning_rate=0.1,
            maximal_learning_rate=1.0,
            step_size=5,
        )
        scheduler = [lrs(i) for i in range(25)]
        self.assertTrue(self.compare_lr_values(scheduler, gold))

    def test_cyclic_tri2_lr(self):
        """Test a triangular2 cyclic learning rate"""
        gold = [0.1,0.28000003,0.45999995,0.64000005,0.81999993,1.0,0.81999993,0.64000005,0.45999995,0.28000003,0.1,
                0.19000003,0.28000003,0.36999995,0.45999995,0.55,0.45999995,0.36999995,0.28000003,0.19000003,0.1,
                0.14499995,0.19000003,0.23499998,0.28000003]
        lrs = cyclic_learning_rate.Triangular2CyclicalLearningRate(
            initial_learning_rate=0.1,
            maximal_learning_rate=1.0,
            step_size=5,
        )
        scheduler = [lrs(i) for i in range(25)]
        self.assertTrue(self.compare_lr_values(scheduler, gold))

    def test_cyclic_exp_lr(self):
        """Test a exponential cyclic learning rate"""
        gold = [0.1,0.26200002431869507,0.39159995317459106,0.49366000294685364,0.572391927242279,0.6314409971237183,
                0.4826374351978302,0.35828033089637756,0.25496816635131836,0.16973569989204407,0.10000000149011612,
                0.156485915184021,0.2016746699810028,0.23726072907447815,0.2647128999233246,0.2853020131587982,
                0.23341745138168335,0.19005677103996277,0.1540340781211853,0.12431533634662628,0.10000000149011612,
                0.11969540268182755,0.1354517638683319,0.1478598564863205,0.15743185579776764]

        lrs = cyclic_learning_rate.ExponentialCyclicalLearningRate(
            initial_learning_rate=0.1,
            maximal_learning_rate=1.0,
            step_size=5,
            gamma=0.9,
        )
        scheduler = [lrs(i) for i in range(25)]
        self.assertTrue(self.compare_lr_values(scheduler, gold))

    def test_exp_lr(self):
        """Test a exponential learning rate"""
        gold = [0.10000000149011612,0.09486832469701767,0.08999999612569809,0.08538150042295456,0.08099999278783798,
                0.07684334367513657,0.07289998978376389,0.06915900856256485,0.06560999155044556,0.062243103981018066,
                0.05904899165034294,0.05601879581809044,0.05314409360289574,0.050416912883520126,0.04782968387007713,
                0.045375220477581024,0.04304671287536621,0.04083769768476486,0.03874203935265541,0.036753926426172256,
                0.03486783429980278,0.03307853266596794,0.03138105198740959,0.029770677909255028,0.028242945671081543]

        lrs = ExponentialDecay(
            initial_learning_rate=0.1,
            decay_steps=2,
            decay_rate=0.9,
        )
        scheduler = [lrs(i) for i in range(25)]
        self.assertTrue(self.compare_lr_values(scheduler, gold))

    def test_constant_lr(self):
        """Test a constant learning rate"""
        lrs = ExponentialDecay(
            initial_learning_rate=0.1,
            decay_steps=10000000,
            decay_rate=1.0,
        )
        for i in range(25):
            assert round(float(lrs(i).numpy()), 5) == 0.1

    def test_cyclic_lr_in_training_pipeline(self):
        """Test a cyclic learning rate in model training"""
        Logger = logging_utils.setup_logging(
            reset=True,
            file_name=os.path.join(INPUT_DIR + 'ranklib', "output_log.csv"),
            log_to_file=True,
        )

        io = LocalIO()
        feature_config = self.parse_config(TFRecordTypeKey.SEQUENCE_EXAMPLE, self.feature_config_yaml_convert_to_clicks,
                                           io)

        dataset = RelevanceDataset(
            data_dir=INPUT_DIR + '/ranklib',
            data_format=DataFormatKey.RANKLIB,
            feature_config=feature_config,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            batch_size=2,
            file_io=io,
            preprocessing_keys_to_fns={},
            logger=Logger,
            keep_additional_info=KEEP_ADDITIONAL_INFO,
            non_zero_features_only=NON_ZERO_FEATURES_ONLY,
            max_sequence_size=319,
        )

        # Define interaction model
        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=feature_config,
            feature_layer_keys_to_fns={},
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            max_sequence_size=319,
            file_io=io,
        )

        # Define loss object from loss key
        loss: RelevanceLossBase = loss_factory.get_loss(
            loss_key=LossKey.RANK_ONE_LISTNET, scoring_type=ScoringTypeKey.POINTWISE
        )

        # Define scorer
        scorer: ScorerBase = RelevanceScorer.from_model_config_file(
            model_config_file=self.model_config_file,
            interaction_model=interaction_model,
            loss=loss,
            logger=Logger,
            file_io=io,
        )

        optimizer: Optimizer = get_optimizer(
            model_config=io.read_yaml(self.model_config_file))

        # Combine the above to define a RelevanceModel
        relevance_model: RelevanceModel = RankingModel(
            feature_config=feature_config,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            scorer=scorer,
            optimizer=optimizer,
            model_file=None,
            file_io=io,
            logger=Logger,
        )
        callbacks_list = []
        my_callback_object = LrCallback()
        callbacks_list.append(my_callback_object)

        history = relevance_model.model.fit(
            x=dataset.train,
            validation_data=dataset.validation,
            epochs=2,
            verbose=True,
            callbacks=callbacks_list,
        )
        lr_list = my_callback_object.get_lr_list()
        lr_gold = [0.001,0.020800006,0.040599994,0.0604,0.080199994,0.1,0.080199994,0.0604,0.040599994,0.020800006,0.001,
                   0.010900003,0.020800006,0.030699994,0.040599994,0.050499998,0.040599994,0.030699994,0.020800006,
                   0.010900003,0.001,0.0059499955,0.010900003,0.015849996,0.020800006,0.02575,0.020800006,0.015849996,
                   0.010900003,0.0059499955,0.001,0.0034749978,0.0059500015,0.008424998,0.010900003,0.013375,
                   0.010900003,0.008424998,0.0059500015,0.0034749978,0.001,0.0022374988,0.0034749978,0.0047125025,
                   0.0059500015,0.0071875,0.0059500015,0.0047125025,0.0034749978, 0.0022374988]

        assert np.all(np.isclose(lr_gold, lr_list))

    def test_reduce_lr_on_plateau_in_training_pipeline(self):
        """Test reduce lr on plateau"""
        self.model_config_file = MODEL_CONFIG_REDUCE_LR_ON_PLATEAU
        Logger = logging_utils.setup_logging(
            reset=True,
            file_name=os.path.join(INPUT_DIR + 'ranklib', "output_log.csv"),
            log_to_file=True,
        )

        io = LocalIO()
        feature_config = self.parse_config(TFRecordTypeKey.SEQUENCE_EXAMPLE, self.feature_config_yaml_convert_to_clicks,
                                           io)
        model_config = io.read_yaml(self.model_config_file)

        dataset = RelevanceDataset(
            data_dir=INPUT_DIR + '/ranklib',
            data_format=DataFormatKey.RANKLIB,
            feature_config=feature_config,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            batch_size=32,
            file_io=io,
            preprocessing_keys_to_fns={},
            logger=Logger,
            keep_additional_info=KEEP_ADDITIONAL_INFO,
            non_zero_features_only=NON_ZERO_FEATURES_ONLY,
            max_sequence_size=319,
        )

        # Define interaction model
        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=feature_config,
            feature_layer_keys_to_fns={},
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            max_sequence_size=319,
            file_io=io,
        )

        # Define loss object from loss key
        loss: RelevanceLossBase = loss_factory.get_loss(
            loss_key=LossKey.RANK_ONE_LISTNET, scoring_type=ScoringTypeKey.POINTWISE
        )

        # Define scorer
        scorer: ScorerBase = RelevanceScorer.from_model_config_file(
            model_config_file=self.model_config_file,
            interaction_model=interaction_model,
            loss=loss,
            logger=Logger,
            file_io=io,
        )

        optimizer: Optimizer = get_optimizer(model_config=model_config)

        # Combine the above to define a RelevanceModel
        relevance_model: RelevanceModel = RankingModel(
            feature_config=feature_config,
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            scorer=scorer,
            optimizer=optimizer,
            model_file=None,
            file_io=io,
            logger=Logger,
        )
        callback_list = []
        callback_list.append(relevance_model.define_scheduler_as_callback(None, model_config))
        my_callback_object = LrCallback()
        callback_list.append(my_callback_object)

        history = relevance_model.model.fit(
            x=dataset.train.shard(2, 0),
            validation_data=dataset.validation.shard(2, 1),
            epochs=10,
            verbose=True,
            callbacks=callback_list,
        )
        lr_list = my_callback_object.get_lr_reduce_on_plateau_list()
        lr_gold = [50.0, 50.0, 25.0, 12.5, 6.25, 3.125, 1.5625, 1.0, 1.0, 1.0]

        assert np.all(np.isclose(lr_gold, lr_list))


if __name__ == "__main__":
    unittest.main()



"""Tests for calibration implemented in `ml4ir.base.model.calibrations.temperature_scaling` """
import shutil
import os

import pandas as pd
import tensorflow as tf
import numpy as np

from ml4ir.base.model.calibration.temperature_scaling import dict_to_csv, \
    TEMPERATURE_SCALE, accuracy, get_logits_labels, temperature_scale, get_intermediate_model
from ml4ir.applications.classification.tests.test_base import ClassificationTestBase

TEMPERATURE_LAYER_NAME = 'temperature_layer'


class TestCalibration(ClassificationTestBase):
    """Class to test temperature scaling from `ml4ir.base.model.calibration.temperature_scaling` """

    def test_dict_to_csv(self):
        """Tests if the .zip file has been created and contains .csv file.
         It also tests if .csv file contains true values"""

        data_dict = {'feature': [1, 2.0, 3], 'labels': [0, 0, 1]}
        dict_to_csv(data_dict, self.output_dir, self.file_io, zip_output=True)

        filename_zip = os.path.join(self.output_dir, TEMPERATURE_SCALE)
        filename_csv = os.path.join(self.output_dir, TEMPERATURE_SCALE, f'{TEMPERATURE_SCALE}.csv')

        shutil.unpack_archive(f'{filename_zip}.zip', self.output_dir, "zip")
        self.assertTrue(os.path.exists(filename_csv))

        df = pd.read_csv(filename_csv)
        pd.testing.assert_frame_equal(df, pd.DataFrame.from_dict(data_dict))

    def test_accuracy(self):
        """Tests accuracy"""
        logits = np.array([[0.3, 0.2, 0.5], [0.7, 0.2, 0.1]])
        labels = np.array([2, 1])
        scores_tensor = tf.constant(logits, name='logits_test', dtype=tf.float32)
        labels_tensor = tf.constant(labels, name='logits_test', dtype=tf.float32)

        acc = accuracy(scores_tensor, labels_tensor)
        self.assertEqual(acc, 0.5, msg="accuracy function does not work as expected")

    def test_temperature_scaling(self):
        """Tests temperature scaling """

        # Computing logits before temperature scaling of the validation set
        logits_numpys, labels_numpys = get_logits_labels(self.classification_model.model,
                                                         self.relevance_dataset.validation)
        # Sanity check the shape of logits and labels
        self.assertTrue(len(logits_numpys) == len(labels_numpys))

        results = temperature_scale(self.classification_model.model,
                                    self.classification_model.scorer,
                                    self.relevance_dataset,
                                    self.logger,
                                    self.output_dir, 1.5, self.file_io)

        # Tests the learned temperature parameter
        expected_value = 1.55
        atol = 0.1
        self.assertTrue(np.isclose(results.position[0], expected_value, atol=atol))

        # computing logits after temperature scaling of the validation set
        logits_numpys_w_ts, labels_numpys_w_ts = get_logits_labels(self.classification_model.model,
                                                                   self.relevance_dataset.validation)

        # Tests if the model weights were frozen and not affected by temperature scaling
        np.testing.assert_array_equal(logits_numpys, logits_numpys_w_ts)
        np.testing.assert_array_equal(labels_numpys, labels_numpys_w_ts)

        # Tests the accuracy of the validation set before and after temperature scaling
        logits_tensor = tf.constant(logits_numpys, dtype='float32')
        labels_tenosr = tf.constant(labels_numpys, dtype='int32')
        acc_org = accuracy(logits_tensor, labels_tenosr)

        logits_w_ts_tensor = tf.constant(logits_numpys_w_ts, dtype='float32')
        labels_w_ts_tenosr = tf.constant(labels_numpys_w_ts, dtype='int32')
        acc_ts = accuracy(logits_w_ts_tensor, labels_w_ts_tenosr)

        np.testing.assert_array_equal(acc_org, acc_ts, err_msg="the accuracy of the validation "
                                                               "set before and after temperature "
                                                               "scaling differs")

    def test_add_temperature_layer(self):
        """Tests whether adding temperature scaling layer scales the logits as expected """

        # Tests whether temperature = 1.0 has no effect on outputs
        output_original = self.classification_model.model.predict(
            self.relevance_dataset.validation).squeeze()

        temperature = 1.0
        self.classification_model.add_temperature_layer(temperature=temperature,
                                                        layer_name=TEMPERATURE_LAYER_NAME)

        output_calibration = self.classification_model.model.predict(
            self.relevance_dataset.validation).squeeze()

        rtol = 1e-6
        np.testing.assert_allclose(output_original, output_calibration, rtol=rtol,
                                   err_msg="outputs differs between models with and "
                                           "without temperature = 1.0 ")

        # refreshing relevance model
        self.classification_model = self.classification_pipeline.get_relevance_model()

        # getting logits of val. set from original model (without temperature scaling)
        model_wo_ts = get_intermediate_model(self.classification_model.model,
                                                  self.classification_model.scorer)
        logits_numpys_wo_ts, labels_numpys_wo_ts = get_logits_labels(model_wo_ts,
                                                                     self.relevance_dataset.
                                                                     validation)

        # adding a temperature scaling layer to the original model
        temperature = 1.5
        self.classification_model.add_temperature_layer(temperature=temperature,
                                                        layer_name=TEMPERATURE_LAYER_NAME)

        # getting logits of val. set from the model with temperature scaling layer
        temperature_layer = self.classification_model.model.get_layer(name=TEMPERATURE_LAYER_NAME)
        model_wth_ts = tf.keras.models.Model(self.classification_model.model.input,
                                             temperature_layer.output)

        logits_numpys_w_ts, labels_numpys_w_ts = get_logits_labels(model_wth_ts,
                                                                   self.relevance_dataset.validation)

        # tests if the scaled logits w.r.t. temperature are the same as the  model with temperature
        # scaling layer logits
        rtol = 1e-6
        np.testing.assert_allclose(logits_numpys_wo_ts/temperature, logits_numpys_w_ts,
                                   rtol=rtol, err_msg="scaled logits w.r.t. temperature are not "
                                                      "the same as the model with TS layer logits")

    def test_relevance_model_w_ts_save(self):
        """Tests whether a loaded model with temperature scaling layer predicts same output with
        the initial model """

        # refreshing relevance model
        self.classification_model = self.classification_pipeline.get_relevance_model()
        initial_model_no_layers = len(self.classification_model.model.layers)

        # adding Temperature Scaling layer
        temperature = 1.5
        self.classification_model.add_temperature_layer(temperature=temperature,
                                                        layer_name=TEMPERATURE_LAYER_NAME)
        model_w_temperature_no_layers = len(self.classification_model.model.layers)

        # tests whether the model with TS layer has one additional (TS) layer
        self.assertEqual(initial_model_no_layers + 1, model_w_temperature_no_layers,
                         msg="difference between number of layers of models with and without "
                             "temperature should be one")

        outputs_before_saving = self.classification_model.model.predict(
            self.relevance_dataset.validation).squeeze()

        # saving and loading the model with TS layer
        self.classification_model.save(
            models_dir=self.output_dir,
            preprocessing_keys_to_fns={},
            postprocessing_fn=None,
            required_fields_only=not self.args.use_all_fields_at_inference,
            pad_sequence=self.args.pad_sequence_at_inference,
            sub_dir='final_calibrated'
        )
        # refreshing relevance model and loading the saved model
        # currently the TS layer is not part of architecture building and needs to be added to
        # the model separately
        self.classification_model = self.classification_pipeline.get_relevance_model()
        # deliberately assigning a different temperature value here
        self.classification_model.add_temperature_layer(temperature=1.0)
        path = os.path.join(self.output_dir, 'final_calibrated', 'default')
        self.classification_model.load_weights(path)

        # tests whether the model before saving and loaded model have equal number of layers
        self.assertEqual(model_w_temperature_no_layers,
                         len(self.classification_model.model.layers),
                         msg="number of layers for model with temperature before and after "
                             "saving differs!")

        ouputs_loaded_model = self.classification_model.model.predict(
            self.relevance_dataset.validation).squeeze()

        # tests whether the saved model and loaded model predicts same results
        rtol = 1e-6
        np.testing.assert_allclose(outputs_before_saving, ouputs_loaded_model, rtol=rtol,
                                   err_msg="Outputs differ between models with temperature before"
                                           " and after saving!")







import os
import pandas as pd
import tensorflow as tf

from ml4ir.base.io.file_io import FileIO
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.data import tfrecord_reader, tfrecord_writer, ranklib_helper

from typing import List

TFRECORD_FILE = "file_0.tfrecord"

def read(
    data_dir: str,
    feature_config: FeatureConfig,
    tfrecord_type: str,
    tfrecord_dir: str,
    file_io: FileIO,
    batch_size: int = 128,
    preprocessing_keys_to_fns: dict = {},
    use_part_files: bool = False,
    max_sequence_size: int = 25,
    parse_tfrecord: bool = True,
    logger=None,
    keep_additional_info: bool = False,
    non_zero_features_only: bool = True,
    **kwargs
) -> tf.data.TFRecordDataset:
    """
    - reads ranklib-formatted data from an input directory
    - selects relevant features
    - creates Dataset X and y

    Current execution plan:
        1. Convert ranklib to a dataframe
        2. Convert each query into tf.train.SequenceExample protobufs
        3. Write the protobufs into a .tfrecord file
        4. Load .tfrecord file into a TFRecordDataset and parse the protobufs

        Parameters
        ----------
        data_dir: str
            Path to directory containing csv files to read
        feature_config: ml4ir.config.features.FeatureConfig object
            FeatureConfig object extracted from the feature config
        tfrecord_dir: str
            Path to directory where the serialized .tfrecord files will be stored
        batch_size: int
            Value specifying the size of the batch
        use_part_files: bool
            Value specifying whether to look for part files
        max_sequence_size: int
            Value specifying max number of records per query
        logger: logging object
            logging object
        keep_additional_info: int
            Option to keep additional info (All info after the "#") 1 to keep, 0 to ignore
        non_zero_features_only: int
            Only non zero features are stored. 1 for yes, 0 otherwise

        Returns
        -------
        tensorflow TFRecordDataset
            Processed dataset
    """
    ranklib_files: List[str] = file_io.get_files_in_directory(
        data_dir,
        extension="" if use_part_files else ".txt",
        prefix="part-" if use_part_files else "",
    )

    gl_2_clicks = False

    # Create a directory for storing tfrecord files
    file_io.make_directory(tfrecord_dir, clear_dir=True)

    #Convert input ranklib file to dataframe
    df = pd.concat(
        [
            ranklib_helper.convert(f, keep_additional_info, gl_2_clicks, non_zero_features_only,
                                   feature_config.get_query_key()['name'], feature_config.get_label()['name'])
            for f in ranklib_files
        ])

    #Write tfrecord files
    tfrecord_writer.write_from_df(df=df,
                                  tfrecord_file=os.path.join(tfrecord_dir, TFRECORD_FILE),
                                  feature_config=feature_config,
                                  tfrecord_type=tfrecord_type,
                                  logger=logger)


    dataset = tfrecord_reader.read(
        data_dir=tfrecord_dir,
        feature_config=feature_config,
        tfrecord_type=tfrecord_type,
        max_sequence_size=max_sequence_size,
        batch_size=batch_size,
        preprocessing_keys_to_fns=preprocessing_keys_to_fns,
        parse_tfrecord=parse_tfrecord,
        file_io=file_io,
        logger=logger,
    )

    return dataset


import os
import sys
import tensorflow as tf
from tensorflow import data
import pandas as pd
import numpy as np
from typing import Optional

from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.model.scoring.prediction_helper import get_predict_fn
from ml4ir.base.model.relevance_model import RelevanceModelConstants
from ml4ir.applications.ranking.model.scoring import prediction_helper
from ml4ir.applications.ranking.model.metrics import metrics_helper
from ml4ir.applications.ranking.config.keys import PositionalBiasHandler
from ml4ir.applications.ranking.t_test import perform_click_rank_dist_paired_t_test, compute_stats_from_stream, t_test_log_results


pd.set_option("display.max_rows", 500)
pd.set_option("display.max_columns", 500)


class RankingConstants:
    NEW_RANK = "new_rank"
    TTEST_PVALUE_THRESHOLD = 0.1


class RankingModel(RelevanceModel):
    def predict(
        self,
        test_dataset: data.TFRecordDataset,
        inference_signature: str = "serving_default",
        additional_features: dict = {},
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
    ):
        """
        Predict the scores on the test dataset using the trained model

        Parameters
        ----------
        test_dataset : `Dataset` object
            `Dataset` object for which predictions are to be made
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status

        Returns
        -------
        `pd.DataFrame`
            pandas DataFrame containing the predictions on the test dataset
            made with the `RelevanceModel`
        """
        additional_features[RankingConstants.NEW_RANK] = prediction_helper.convert_score_to_rank

        return super().predict(
            test_dataset=test_dataset,
            inference_signature=inference_signature,
            additional_features=additional_features,
            logs_dir=logs_dir,
            logging_frequency=logging_frequency,
        )


    def run_ttest(self, mean, variance, n, ttest_pvalue_threshold):
        """
        Compute the paired t-test statistic and its p-value given mean, standard deviation and sample count
        Parameters
        ----------
        mean: float
            The mean of the rank differences for the entire dataset
        variance: float
            The variance of the rank differences for the entire dataset
        n: int
            The number of samples in the entire dataset
        ttest_pvalue_threshold: float
            P-value threshold for student t-test

        Returns
        -------
        t_test_metrics_dict: Dictionary
            A dictionary with the t-test metrics recorded.
        """
        t_test_stat, pvalue = perform_click_rank_dist_paired_t_test(mean, variance, n)
        t_test_log_results(t_test_stat, pvalue, ttest_pvalue_threshold, self.logger)

        # Log t-test statistic and p-value
        t_test_metrics_dict = {'Rank distribution t-test statistic': t_test_stat,
                               'Rank distribution t-test pvalue': pvalue,
                               'Rank distribution t-test difference is statistically significant': pvalue < ttest_pvalue_threshold}
        return t_test_metrics_dict

    def evaluate(
        self,
        test_dataset: data.TFRecordDataset,
        inference_signature: str = None,
        additional_features: dict = {},
        group_metrics_min_queries: int = 50,
        logs_dir: Optional[str] = None,
        logging_frequency: int = 25,
        compute_intermediate_stats: bool = True,
    ):
        """
        Evaluate the RelevanceModel

        Parameters
        ----------
        test_dataset: an instance of tf.data.dataset
        inference_signature : str, optional
            If using a SavedModel for prediction, specify the inference signature to be used for computing scores
        additional_features : dict, optional
            Dictionary containing new feature name and function definition to
            compute them. Use this to compute additional features from the scores.
            For example, converting ranking scores for each document into ranks for
            the query
        group_metrics_min_queries : int, optional
            Minimum count threshold per group to be considered for computing
            groupwise metrics
        logs_dir : str, optional
            Path to directory to save logs
        logging_frequency : int
            Value representing how often(in batches) to log status
        compute_intermediate_stats : bool
            [Currently ignored] Determines if group metrics and other intermediate stats on the test set should be computed

        Returns
        -------
        df_overall_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing overall metrics
        df_groupwise_metrics : `pd.DataFrame` object
            `pd.DataFrame` containing groupwise metrics if
            group_metric_keys are defined in the FeatureConfig
        metrics_dict : dict
            metrics as a dictionary of metric names mapping to values

        Notes
        -----
        You can directly do a `model.evaluate()` only if the keras model is compiled

        Override this method to implement your own evaluation metrics.
        """
        metrics_dict = dict()
        group_metrics_keys = self.feature_config.get_group_metrics_keys()
        evaluation_features = (
            group_metrics_keys
            + [
                self.feature_config.get_query_key(),
                self.feature_config.get_label(),
                self.feature_config.get_rank(),
            ]
            + [
                f
                for f in self.feature_config.get_secondary_labels()
                if f.get(
                    "node_name",
                    f["name"] not in self.feature_config.get_group_metrics_keys(
                        "node_name"),
                )
            ]
        )
        additional_features[RankingConstants.NEW_RANK] = prediction_helper.convert_score_to_rank

        _predict_fn = get_predict_fn(
            model=self.model,
            tfrecord_type=self.tfrecord_type,
            feature_config=self.feature_config,
            inference_signature=inference_signature,
            is_compiled=self.is_compiled,
            output_name=self.output_name,
            features_to_return=evaluation_features,
            additional_features=additional_features,
            max_sequence_size=self.max_sequence_size,
        )

        batch_count = 0
        df_grouped_stats = pd.DataFrame()
        # defining variables to compute running mean and variance for t-test computations
        agg_count, agg_mean, agg_M2 = 0, 0, 0
        for predictions_dict in test_dataset.map(_predict_fn).take(-1):
            predictions_df = pd.DataFrame(predictions_dict)

            # Accumulating statistics for t-test calculation using 1/rank
            clicked_records = predictions_df[predictions_df[self.feature_config.get_label("node_name")] == 1.0]
            diff = (1/clicked_records[RankingConstants.NEW_RANK] - 1/clicked_records[
                self.feature_config.get_rank("node_name")]).to_list()
            agg_count, agg_mean, agg_M2 = compute_stats_from_stream(diff, agg_count, agg_mean, agg_M2)

            df_batch_grouped_stats = metrics_helper.get_grouped_stats(
                df=predictions_df,
                query_key_col=self.feature_config.get_query_key("node_name"),
                label_col=self.feature_config.get_label("node_name"),
                old_rank_col=self.feature_config.get_rank("node_name"),
                new_rank_col=RankingConstants.NEW_RANK,
                group_keys=list(set(self.feature_config.get_group_metrics_keys(
                    "node_name"))),
                secondary_labels=list(set(self.feature_config.get_secondary_labels(
                    "node_name"))),
            )
            if df_grouped_stats.empty:
                df_grouped_stats = df_batch_grouped_stats
            else:
                df_grouped_stats = df_grouped_stats.add(
                    df_batch_grouped_stats, fill_value=0.0)
            batch_count += 1
            if batch_count % logging_frequency == 0:
                self.logger.info(
                    "Finished evaluating {} batches".format(batch_count))

        # performing click rank distribution t-test
        t_test_metrics_dict = self.run_ttest(agg_mean, (agg_M2 / (agg_count - 1)), agg_count, RankingConstants.TTEST_PVALUE_THRESHOLD)
        metrics_dict.update(t_test_metrics_dict)

        # Compute overall metrics
        df_overall_metrics = metrics_helper.summarize_grouped_stats(
            df_grouped_stats)
        self.logger.info("Overall Metrics: \n{}".format(df_overall_metrics))

        # Log metrics to weights and biases
        metrics_dict.update(
            {"test_{}".format(k): v for k,
             v in df_overall_metrics.to_dict().items()}
        )

        df_group_metrics = None
        df_group_metrics_summary = None
        if group_metrics_keys:
            # Filter groups by min_query_count
            df_grouped_stats = df_grouped_stats[
                df_grouped_stats["query_count"] >= group_metrics_min_queries
            ]

            # Compute group metrics
            df_group_metrics = df_grouped_stats.apply(
                metrics_helper.summarize_grouped_stats, axis=1
            )
            if logs_dir:
                self.file_io.write_df(
                    df_group_metrics,
                    outfile=os.path.join(
                        logs_dir, RelevanceModelConstants.GROUP_METRICS_CSV_FILE),
                )

            # Compute group metrics summary
            df_group_metrics_summary = df_group_metrics.describe()
            self.logger.info(
                "Computing group metrics using keys: {}".format(
                    self.feature_config.get_group_metrics_keys("node_name")
                )
            )
            self.logger.info("Groupwise Metrics: \n{}".format(
                df_group_metrics_summary.T))

            # Log metrics to weights and biases
            metrics_dict.update(
                {
                    "test_group_mean_{}".format(k): v
                    for k, v in df_group_metrics_summary.T["mean"].to_dict().items()
                }
            )

        return df_overall_metrics, df_group_metrics, metrics_dict

    def save(
        self,
        models_dir: str,
        preprocessing_keys_to_fns={},
        postprocessing_fn=None,
        required_fields_only: bool = True,
        pad_sequence: bool = False,
        dataset: Optional[RelevanceDataset] = None,
        experiment_details: Optional[dict] = None
    ):
        """
        Save the RelevanceModel as a tensorflow SavedModel to the `models_dir`
        Additionally, sets the score for the padded records to 0

        There are two different serving signatures currently used to save the model
            `default`: default keras model without any pre/post processing wrapper
            `tfrecord`: serving signature that allows keras model to be served using TFRecord proto messages.
                      Allows definition of custom pre/post processing logic

        Additionally, each model layer is also saved as a separate numpy zipped
        array to enable transfer learning with other ml4ir models.

        Parameters
        ----------
        models_dir : str
            path to directory to save the model
        preprocessing_keys_to_fns : dict
            dictionary mapping function names to tf.functions that should be saved in the preprocessing step of the tfrecord serving signature
        postprocessing_fn: function
            custom tensorflow compatible postprocessing function to be used at serving time.
            Saved as part of the postprocessing layer of the tfrecord serving signature
        required_fields_only: bool
            boolean value defining if only required fields
            need to be added to the tfrecord parsing function at serving time
        pad_sequence: bool, optional
            Value defining if sequences should be padded for SequenceExample proto inputs at serving time.
            Set this to False if you want to not handle padded scores.
        dataset : `RelevanceDataset` object
            RelevanceDataset object that can optionally be passed to be used by downstream jobs
            that want to save the data along with the model.
            Note that this feature is currently unimplemented and is upto the users to override
            and customize.
        experiment_details: dict
            Dictionary containing metadata and results about the current experiment

        Notes
        -----
        All the functions passed under `preprocessing_keys_to_fns` here must be
        serializable tensor graph operations
        """

        def mask_padded_records(predictions, features_dict):
            for key, value in predictions.items():
                predictions[key] = tf.where(
                    tf.equal(features_dict["mask"], 0), tf.constant(
                        0.0), predictions[key]
                )

            return predictions

        super().save(
            models_dir=models_dir,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            postprocessing_fn=mask_padded_records,
            required_fields_only=required_fields_only,
            pad_sequence=pad_sequence,
        )

        # Logging positional biases
        for layer in self.model.layers:
            if layer.name == PositionalBiasHandler.FIXED_ADDITIVE_POSITIONAL_BIAS:
                positional_bias_coefficients = pd.DataFrame(
                    [{'rank': i + 1, 'positional_bias': layer.get_weights()[0][i][0]}
                     for i in range(len(layer.get_weights()[0]))]
                )
                positional_biases = positional_bias_coefficients['positional_bias']
                softmax_positional_biases = np.exp(positional_biases) / np.sum(np.exp(positional_biases), axis=0)
                positional_bias_coefficients['softmax_positional_bias'] = softmax_positional_biases
                self.file_io.write_df(
                    positional_bias_coefficients,
                    outfile=os.path.join(models_dir, "positional_biases.csv"),
                    index=False
                )


class LinearRankingModel(RankingModel):
    """
    Subclass of the RankingModel object customized for training and saving a
    simple linear ranking model. Current implementation overrides the save model
    functionality to map input feature nodes into corresponding weights/coefficients.
    """

    def save(
        self,
        models_dir: str,
        preprocessing_keys_to_fns={},
        postprocessing_fn=None,
        required_fields_only: bool = True,
        pad_sequence: bool = False,
        dataset: Optional[RelevanceDataset] = None,
        experiment_details: Optional[dict] = None
    ):
        """
        Save the RelevanceModel as a tensorflow SavedModel to the `models_dir`
        Additionally, sets the score for the padded records to 0

        There are two different serving signatures currently used to save the model
            `default`: default keras model without any pre/post processing wrapper
            `tfrecord`: serving signature that allows keras model to be served using TFRecord proto messages.
                      Allows definition of custom pre/post processing logic

        Parameters
        ----------
        models_dir : str
            path to directory to save the model
        preprocessing_keys_to_fns : dict
            dictionary mapping function names to tf.functions that should be saved in the preprocessing step of the tfrecord serving signature
        postprocessing_fn: function
            custom tensorflow compatible postprocessing function to be used at serving time.
            Saved as part of the postprocessing layer of the tfrecord serving signature
        required_fields_only: bool
            boolean value defining if only required fields
            need to be added to the tfrecord parsing function at serving time
        pad_sequence: bool, optional
            Value defining if sequences should be padded for SequenceExample proto inputs at serving time.
            Set this to False if you want to not handle padded scores.
        dataset : `RelevanceDataset` object
            RelevanceDataset object that can optionally be passed to be used by downstream jobs
            that want to save the data along with the model.
            Note that this feature is currently unimplemented and is upto the users to override
            and customize.
        experiment_details: dict
            Dictionary containing metadata and results about the current experiment

        Notes
        -----
        - All the functions passed under `preprocessing_keys_to_fns` here must be
        serializable tensor graph operations
        - The LinearRankingModel.save() method specifically saves the weights
        of the dense layer as a CSV file where the feature names are mapped to
        the weights as key-value pairs.
        """

        # Save the linear model coefficients as a CSV
        dense_layer = None
        for layer in self.model.layers:
            if isinstance(layer, tf.keras.layers.Dense):
                dense_layer = layer
                assert dense_layer.units == 1
        if not dense_layer:
            raise KeyError("No dense layer found in the model. Can not save the linear ranking model coefficients.")

        linear_model_coefficients = pd.DataFrame(
            list(zip(
                [f.name.split(":")[0].replace("_expanded", "") for f in self.model.get_layer(
                    "tf_op_layer_train_features").input],
                tf.squeeze(dense_layer.get_weights()[0]).numpy())
            ),
            columns=["feature", "weight"])
        # Adding log for bias value
        if len(dense_layer.get_weights())>1:
            bias_val = dense_layer.get_weights()[1][0]
            linear_model_coefficients.loc[len(linear_model_coefficients.index)] = ['bias', bias_val]
        self.logger.info("Linear Model Coefficients:\n{}".format(
            linear_model_coefficients.to_csv(index=False)))
        self.file_io.write_df(
            linear_model_coefficients,
            outfile=os.path.join(models_dir, "coefficients.csv"),
            index=False
        )

        # Call super save method to persist the SavedModel files
        super().save(
            models_dir=models_dir,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            postprocessing_fn=postprocessing_fn,
            required_fields_only=required_fields_only,
            pad_sequence=pad_sequence,
        )

    def calibrate(self, **kwargs):
        raise NotImplementedError



import unittest
import os
import warnings
from ml4ir.base.data import ranklib_helper
import pandas as pd


warnings.filterwarnings("ignore")

INPUT_FILE = "ml4ir/applications/ranking/tests/data/ranklib/train/sample.txt"
OUTPUT_FILE = "ml4ir/applications/ranking/tests/data/ranklib/train/sample_ml4ir.csv"
QUERY_ID_NAME = 'qid'
RELEVANCE_NAME = 'relevance'
KEEP_ADDITIONAL_INFO = 1
GL_2_CLICKS = 1
NON_ZERO_FEATURES_ONLY = 0


class TestRanklibConversion(unittest.TestCase):
    def setUp(self):
        pass

    def test_conversion(self):
        """Convert ranklib dataset to a csv"""

        ranklib_helper.ranklib_to_csv(INPUT_FILE, OUTPUT_FILE, KEEP_ADDITIONAL_INFO,
                                      GL_2_CLICKS, NON_ZERO_FEATURES_ONLY, QUERY_ID_NAME, RELEVANCE_NAME)
        df = pd.read_csv(OUTPUT_FILE)
        assert QUERY_ID_NAME in df.columns and RELEVANCE_NAME in df.columns
        assert df[QUERY_ID_NAME].nunique() == 49
        if KEEP_ADDITIONAL_INFO == 1:
            assert len(df.columns) >= 138
        else:
            assert len(df.columns) == 138

        if GL_2_CLICKS == 1:
            assert sorted(list(df[RELEVANCE_NAME].unique())) == [0, 1]

    def tearDown(self):
        # Delete output file
        os.remove(OUTPUT_FILE)


if __name__ == "__main__":
    unittest.main()



from setuptools import find_namespace_packages, setup
import yaml
from itertools import chain


optional_requirements_spec = yaml.safe_load(open("optional_requirements.yaml"))


def load_required_dependencies():
    with open("requirements.txt") as f:
        required = f.read().splitlines()

    # Remove optional requirements from required dependencies
    optional_requirements = set(chain(*optional_requirements_spec.values()))
    return [package for package in required if package not in optional_requirements]


def getReadMe():
    with open("README.md", "r") as f:
        long_description = f.read()
    return long_description


setup(
    name="ml4ir",
    packages=find_namespace_packages(include=["ml4ir.*"]),
    version="0.1.9",
    description="Machine Learning libraries for Information Retrieval",
    long_description=getReadMe(),
    long_description_content_type="text/markdown",
    author="Search Relevance, Salesforce",
    author_email="searchrelevancyscrumteam@salesforce.com",
    url="https://www.salesforce.com/",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
    ],
    include_package_data=True,
    license="ASL 2.0",
    python_requires=">=3.7",
    install_requires=load_required_dependencies(),
    extras_require=optional_requirements_spec
)



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow import io
import numpy as np

from ml4ir.base.io.file_io import FileIO


def bytes_sequence_to_encoding_bilstm(feature_tensor, feature_info, file_io: FileIO):
    """
    Encode a string tensor into an encoding.
    Works by converting the string into a bytes sequence and then generating
    a categorical/char embedding for each of the 256 bytes. The char/byte embeddings
    are then combined using a biLSTM

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor that is to be encoded
    feature_info : dict
        Dictionary representing the feature_config for the input feature
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Encoded feature tensor

    Notes
    -----
    Args under `feature_layer_info`:
        max_length : int
            max length of bytes sequence
        embedding_size : int
            dimension size of the embedding;
            if null, then the tensor is just converted to its one-hot representation
        encoding_size : int
            dimension size of the sequence encoding computed using a biLSTM

    The input dimension for the embedding is fixed to 256 because the string is
    converted into a bytes sequence.
    """
    args = feature_info["feature_layer_info"]["args"]

    # Decode string tensor to bytes
    feature_tensor = io.decode_raw(
        feature_tensor, out_type=tf.uint8, fixed_length=args.get("max_length", None),
    )

    feature_tensor = tf.squeeze(feature_tensor, axis=1)
    if "embedding_size" in args:
        char_embedding = layers.Embedding(
            name="{}_bytes_embedding".format(
                feature_info.get("node_name", feature_info.get("name"))
            ),
            input_dim=256,
            output_dim=args["embedding_size"],
            mask_zero=True,
            input_length=args.get("max_length", None),
        )(feature_tensor)
    else:
        char_embedding = tf.one_hot(feature_tensor, depth=256)

    kernel_initializer = args.get("lstm_kernel_initializer", "glorot_uniform")
    encoding = get_bilstm_encoding(
        embedding=char_embedding,
        lstm_units=int(args["encoding_size"] / 2),
        kernel_initializer=kernel_initializer,
    )
    return encoding


def get_bilstm_encoding(embedding, lstm_units, kernel_initializer="glorot_uniform"):
    """
    Convert sequence into encoding by passing through bidirectional LSTM

    Parameters
    ----------
    sequence_tensor : Tensor object
        Sequence tensor with representations for each time step
    lstm_units : int
        Number of units in the LSTM
    kernel_initializer : str
        Any supported tf.keras.initializers
        e.g., 'ones', 'glorot_uniform', 'lecun_normal' ...

    Returns
    -------
    Tensor object
        Encoded feature tensor
    """
    encoding = layers.Bidirectional(
        layers.LSTM(
            units=lstm_units, return_sequences=False, kernel_initializer=kernel_initializer
        ),
        merge_mode="concat",
    )(embedding)
    encoding = tf.expand_dims(encoding, axis=1)
    return encoding


def global_1d_pooling(feature_tensor, feature_info, file_io: FileIO):
    """
    1D pooling to reduce a variable length sequence feature into a scalar
    value. This method optionally allows users to add multiple such pooling
    operations to produce a fixed dimensional feature vector as well.

    Parameters
    ----------
    feature_tensor : Tensor object
        String feature tensor that is to be aggregated/pooled
        Dimensions -> [batch_size, max_sequence_size, max_len]
    feature_info : dict
        Dictionary representing the feature_config for the input feature
    file_io : FileIO object
        FileIO handler object for reading and writing

    Returns
    -------
    Tensor object
        Global pooled/aggregated feature vector

    Notes
    -----
    Args under `feature_layer_info`:
        fns : list of str
            List of string pooling operations that should be applied.
            Must be one of ["sum", "mean", "max", "min", "count_nonzero"]
        padded_val : int/float
            Value to be ignored from the pooling operations.
        masked_max_val : int/float
            Value used to mask the padded values for computing the max and min
            pooling operations. This allows us to ignore these values in the min
            and max pool operations. For example, if all the values in the tensor
            are in [0., 1.], then a masked_max_val of > 1. will make sure we do
            not pick padded values in the min/max pooling ops.
            Default value: 2
    """
    args = feature_info["feature_layer_info"]["args"]
    pooled_tensors = list()

    # Check if at least one pooling fn is specified
    if len(args["fns"]) == 0:
        raise ValueError(
            "At least 1 pooling function should be specified. Found : {}".format(len(args["fns"]))
        )

    # Define the masking value for each pooling op
    DEFAULT_MASKED_MAX_VAL = 2.0
    masked_val_lookup = {
        "sum": 0.0,
        "mean": 0.0,
        "max": - args.get("masked_max_val", DEFAULT_MASKED_MAX_VAL),
        "min": args.get("masked_max_val", DEFAULT_MASKED_MAX_VAL),
        "count_nonzero": 0.0
    }

    padded_val_tensor = None
    if "padded_val" in args:
        padded_val_tensor = tf.constant(args["padded_val"], dtype=feature_tensor.dtype)

    # Apply the 1D global pooling on the last dimension
    for fn in args["fns"]:
        # If padded_val is passed, mask the padded values appropriately
        if "padded_val" in args:
            masked_feature_tensor = tf.where(
                tf.equal(feature_tensor, padded_val_tensor), masked_val_lookup[fn], feature_tensor
            )
        else:
            masked_feature_tensor = feature_tensor

        # Apply each pooling op
        if fn == "sum":
            pooled_tensors.append(tf.math.reduce_sum(masked_feature_tensor, axis=-1))

        elif fn == "mean":
            if "padded_val" in args:
                # NOTE: To avoid division by zero, we set 0 to a small value of 1e-10
                pooled_tensors.append(
                    tf.math.divide(
                        tf.math.reduce_sum(masked_feature_tensor, axis=-1),
                        tf.math.reduce_sum(
                            tf.where(
                                tf.equal(feature_tensor, padded_val_tensor),
                                tf.constant(1e-10, dtype=feature_tensor.dtype),
                                tf.constant(1, dtype=feature_tensor.dtype),
                            ),
                            axis=-1,
                        ),
                    )
                )
            else:
                pooled_tensors.append(tf.math.reduce_mean(masked_feature_tensor, axis=-1))

        elif fn == "max":
            pooled_tensors.append(tf.math.reduce_max(masked_feature_tensor, axis=-1))

        elif fn == "min":
            pooled_tensors.append(tf.math.reduce_min(masked_feature_tensor, axis=-1))

        elif fn == "count_nonzero":
            pooled_tensors.append(
                tf.cast(
                    tf.math.count_nonzero(masked_feature_tensor, axis=-1),
                    dtype=masked_feature_tensor.dtype,
                )
            )

        else:
            raise KeyError(
                "{} pooling function not supported. Please use one of sum, mean, max, min, count_nonzero.".format(
                    fn
                )
            )

    # Stack all the pooled tensors to get one fixed dimensional dense vector
    return tf.stack(
        pooled_tensors,
        axis=-1,
        name="global_1d_pooling_{}".format(
            feature_info.get("node_name", feature_info.get("name"))
        ),
    )



import tensorflow as tf

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import SequenceExampleTypeKey
from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.features.feature_fns.sequence import bytes_sequence_to_encoding_bilstm
from ml4ir.base.features.feature_fns.sequence import global_1d_pooling
from ml4ir.base.features.feature_fns.categorical import categorical_embedding_to_encoding_bilstm
from ml4ir.base.features.feature_fns.categorical import categorical_embedding_with_hash_buckets
from ml4ir.base.features.feature_fns.categorical import categorical_embedding_with_indices
from ml4ir.base.features.feature_fns.categorical import categorical_embedding_with_vocabulary_file
from ml4ir.base.features.feature_fns.categorical import (
    categorical_embedding_with_vocabulary_file_and_dropout,
)
from ml4ir.base.features.feature_fns.tf_native import tf_native_op
from ml4ir.base.io.file_io import FileIO


class FeatureLayerMap:
    """Class defining mapping from keys to feature layer functions"""

    def __init__(self):
        """
        Define ml4ir's predefined feature transformation functions
        """
        self.key_to_fn = {
            bytes_sequence_to_encoding_bilstm.__name__: bytes_sequence_to_encoding_bilstm,
            global_1d_pooling.__name__: global_1d_pooling,
            categorical_embedding_to_encoding_bilstm.__name__: categorical_embedding_to_encoding_bilstm,
            categorical_embedding_with_hash_buckets.__name__: categorical_embedding_with_hash_buckets,
            categorical_embedding_with_indices.__name__: categorical_embedding_with_indices,
            categorical_embedding_with_vocabulary_file.__name__: categorical_embedding_with_vocabulary_file,
            categorical_embedding_with_vocabulary_file_and_dropout.__name__: categorical_embedding_with_vocabulary_file_and_dropout,
            tf_native_op.__name__: tf_native_op
        }

    def add_fn(self, key, fn):
        """
        Add custom new function to the FeatureLayerMap

        Parameters
        ----------
        key : str
            name of the feature transformation function
        fn : tf.function
            tensorflow function that transforms input features
        """

        self.key_to_fn[key] = fn

    def add_fns(self, keys_to_fns_dict):
        """
        Add custom new functions to the FeatureLayerMap

        Parameters
        ----------
        keykeys_to_fns_dict : dict
            Dictionary with name and definition of custom
            tensorflow functions that transform input features
        """
        self.key_to_fn.update(keys_to_fns_dict)

    def get_fns(self):
        """
        Get all feature transformation functions

        Returns
        -------
        dict
            Dictionary of feature transformation functions
        """
        return self.key_to_fn

    def get_fn(self, key):
        """
        Get feature transformation function using the key

        Parameters
        ----------
        key : str
            Name of the feature transformation function to be fetched

        Returns
        -------
        tf.function
            Feature transformation function
        """
        return self.key_to_fn.get(key)

    def pop_fn(self, key):
        """
        Get feature transformation function using the key and remove
        from FeatureLayerMap

        Parameters
        ----------
        key : str
            Name of the feature transformation function to be fetched

        Returns
        -------
        tf.function
            Feature transformation function
        """
        self.key_to_fn.pop(key)


def define_feature_layer(
    feature_config: FeatureConfig,
    tfrecord_type: str,
    feature_layer_map: FeatureLayerMap,
    file_io: FileIO,
):
    """
    Defines a feature layer function that works on keras.Inputs

    Parameters
    ----------
    feature_config : `FeatureConfig` object
        FeatureConfig object that defines the feature transformation
        function to be applied to each feature
    tfrecord_type : {"example", "sequence_example"}
        String defining the TFRecord type of the data
    feature_layer_map : `FeatureLayerMap`
        `FeatureLayerMap` object mapping custom function names to function definition

    Returns
    -------
        tensorflow op that is the feature layer by collecting all the
        feature transformation functions assigned to the respective
        keras.Inputs features as specified by the FeatureConfig

    Notes
    -----
    Use feature_layer_map to define custom functions when using ml4ir as a library
    """

    def feature_layer_op(inputs):
        """
        Apply feature transformation functions and reshape the input
        feature tensors as specified by the FeatureConfig
        """
        train_features = dict()
        metadata_features = dict()

        # Define a dynamic tensor tiling shape
        # NOTE: Can not be hardcoded as we allow for varying sequence_size at inference time
        if tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE:
            train_tile_shape = tf.shape(
                tf.expand_dims(
                    tf.expand_dims(tf.gather(inputs["mask"], indices=0), axis=0), axis=-1
                )
            )
            metadata_tile_shape = tf.shape(
                tf.expand_dims(tf.gather(inputs["mask"], indices=0), axis=0)
            )

        for feature_info in feature_config.get_all_features(include_label=False):
            feature_node_name = feature_info.get("node_name", feature_info["name"])
            feature_layer_info = feature_info["feature_layer_info"]
            feature_tensor = inputs[feature_node_name]

            if "fn" in feature_layer_info:
                feature_fn = feature_layer_map.get_fn(feature_layer_info["fn"])
                if not feature_fn:
                    raise RuntimeError(
                        "Unsupported feature function: {}".format(feature_layer_info["fn"])
                    )
                feature_tensor = feature_fn(
                    feature_tensor=feature_tensor, feature_info=feature_info, file_io=file_io
                )
            elif feature_info["trainable"]:
                # Default feature layer
                feature_tensor = tf.expand_dims(feature_tensor, axis=-1, name="{}_expanded".format(feature_node_name))

            """
            NOTE: If the trainable feature is of type context, then we tile/duplicate
            the values for all examples of the sequence
            """
            if (
                tfrecord_type == TFRecordTypeKey.SEQUENCE_EXAMPLE
                and feature_info["tfrecord_type"] == SequenceExampleTypeKey.CONTEXT
            ):
                if feature_info["trainable"]:
                    feature_tensor = tf.tile(feature_tensor, train_tile_shape)
                else:
                    feature_tensor = tf.tile(feature_tensor, metadata_tile_shape)

            if feature_info["trainable"]:
                train_features[feature_node_name] = tf.cast(feature_tensor, tf.float32)
            else:
                if feature_info["dtype"] == tf.int64:
                    feature_tensor = tf.cast(feature_tensor, tf.float32)
                metadata_features[feature_node_name] = feature_tensor

        return train_features, metadata_features

    return feature_layer_op



from ml4ir.base.model.architectures import architecture_factory


class ScoringBase(object):
    def __init__(self, scoring_key, model_config, loss_type):
        self.scoring_key = scoring_key
        self.model_config = model_config
        self.loss_type = loss_type

        self.architecture_op = architecture_factory.get_architecture(model_config=model_config)

    def get_scoring_fn(self, **kwargs):
        """
        Define a ranking scoring function with specified architecture
        """

        def _scoring_fn(features):
            pass

        return _scoring_fn



import unittest
import tensorflow as tf
import logging

from ml4ir.base.data.tfrecord_reader import TFRecordSequenceExampleParser
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.config.keys import TFRecordTypeKey
from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.features.preprocessing import PreprocessingMap

DATASET_PATH = "ml4ir/applications/ranking/tests/data/tfrecord/train/file_0.tfrecord"
FEATURE_CONFIG_PATH = "ml4ir/applications/ranking/tests/data/configs/feature_config.yaml"
MAX_SEQUENCE_SIZE = 25


class SequenceExampleParserTest(unittest.TestCase):
    """
    Test class for ml4ir.base.data.tfrecord_reader.TFRecordSequenceExampleParser
    """

    def setUp(self):
        file_io = LocalIO()
        logger = logging.getLogger()

        self.dataset = tf.data.TFRecordDataset(DATASET_PATH)
        self.proto = next(iter(self.dataset))
        self.feature_config = FeatureConfig.get_instance(
            tfrecord_type=TFRecordTypeKey.SEQUENCE_EXAMPLE,
            feature_config_dict=file_io.read_yaml(FEATURE_CONFIG_PATH),
            logger=logger,
        )
        self.parser = TFRecordSequenceExampleParser(
            feature_config=self.feature_config,
            preprocessing_map=PreprocessingMap(),
            required_fields_only=False,
            pad_sequence=True,
            max_sequence_size=25,
        )

    def test_features_spec(self):
        """
        Test the feature specification constructed and used to parse the Example proto
        """
        features_spec = self.parser.features_spec

        assert isinstance(features_spec, tuple)
        assert isinstance(features_spec[0], dict)
        assert isinstance(features_spec[1], dict)

        # Check if the feature specification matches with the feature_config
        assert len(set(self.feature_config.get_context_features("name"))) == len(features_spec[0])
        assert len(set(self.feature_config.get_sequence_features("name"))) == len(features_spec[1])

        for feature in self.feature_config.get_context_features("name"):
            assert feature in features_spec[0]
        for feature in self.feature_config.get_sequence_features("name"):
            assert feature in features_spec[1]

    def test_extract_features_from_proto(self):
        """
        Test extraction of features from serialized proto
        """
        context_features, sequence_features = self.parser.extract_features_from_proto(self.proto)

        for feature in self.feature_config.get_context_features("name"):
            assert feature in context_features

            # Test that each extracted feature is a scalar
            assert context_features[feature].shape == ()

        for feature in self.feature_config.get_sequence_features("name"):
            assert feature in sequence_features

            # Test that all features are sparse tensor
            assert isinstance(sequence_features[feature], tf.sparse.SparseTensor)

            feature_tensor = tf.sparse.to_dense(tf.sparse.reset_shape(sequence_features[feature]))

            assert feature_tensor.shape == (1, 2)

        # Assert that there is no mask feature
        assert "mask" not in sequence_features

    def test_get_default_tensor(self):
        """
        Test the default tensor used for missing features
        """
        default_tensor = self.parser.get_default_tensor(
            self.feature_config.get_feature("query_text"), sequence_size=25
        )
        assert default_tensor.shape == ()

        default_tensor = self.parser.get_default_tensor(
            self.feature_config.get_feature("quality_score"), sequence_size=8
        )
        assert default_tensor.shape == (8,)

    def test_get_feature(self):
        """
        Test fetching feature tensor from extracted feature dictionary
        """
        # Checking context features
        feature_tensor = self.parser.get_feature(
            self.feature_config.get_feature("query_text"),
            extracted_features=({"query_text": tf.zeros((3, 4, 6))}, {}),
            sequence_size=10,
        )
        assert feature_tensor.shape == (1, 3, 4, 6)

        # Check missing feature being replaced with default tensor
        feature_tensor = self.parser.get_feature(
            self.feature_config.get_feature("query_text"),
            extracted_features=({}, {}),
            sequence_size=10,
        )
        assert feature_tensor.shape == (1,)

        # Checking sequence features
        feature_tensor = self.parser.get_feature(
            self.feature_config.get_feature("quality_score"),
            extracted_features=({}, {"quality_score": tf.zeros((3, 4, 6))}),
            sequence_size=10,
        )
        assert feature_tensor.shape == (3, 4, 6)

        # Check missing feature being replaced with default tensor
        feature_tensor = self.parser.get_feature(
            self.feature_config.get_feature("quality_score"),
            extracted_features=({}, {}),
            sequence_size=10,
        )
        assert feature_tensor.shape == (10,)

    def test_generate_and_add_mask(self):
        """
        Test mask generation and addition
        """
        rank_tensor = tf.constant([[1, 2, 3, 4, 5]])
        indices = tf.where(tf.not_equal(rank_tensor, tf.constant(0)))
        values = tf.gather_nd(rank_tensor, indices)
        sparse_rank_tensor = tf.SparseTensor(indices, values, rank_tensor.shape)

        # Check when pad sequence is set to True
        features_dict, sequence_size = self.parser.generate_and_add_mask(
            ({}, {"rank": sparse_rank_tensor}), {}
        )

        assert "mask" in features_dict
        assert features_dict["mask"].shape == (25,)
        assert tf.reduce_sum(features_dict["mask"]).numpy() == 5
        assert sequence_size == 25

        # Check when pad sequence is set to False
        self.parser.pad_sequence = False
        features_dict, sequence_size = self.parser.generate_and_add_mask(
            ({}, {"rank": sparse_rank_tensor}), {}
        )

        assert "mask" in features_dict
        assert features_dict["mask"].shape == (5,)
        assert tf.reduce_sum(features_dict["mask"]).numpy() == 5
        assert sequence_size == 5
        self.parser.pad_sequence = True

    def test_pad_feature(self):
        """
        Test feature padding to max sequence size
        """
        # Check no padding for context features
        feature_tensor = self.parser.pad_feature(tf.zeros((10)),
                                                 self.feature_config.get_feature("query_text"))
        assert feature_tensor.shape == (10,)

        # Check padding for sequence features
        feature_tensor = self.parser.pad_feature(tf.zeros((10)),
                                                 self.feature_config.get_feature("quality_score"))
        assert feature_tensor.shape == (25,)

    def test_parse_fn(self):
        """
        Test the Example parsing function
        """
        # Check tensor shapes when pad_sequence is True
        features, labels = self.parser.get_parse_fn()(self.proto)

        assert isinstance(features, dict)
        assert isinstance(labels, tf.Tensor)

        for feature in self.feature_config.get_all_features(key="node_name", include_label=False):
            assert feature in features

        assert features["mask"].shape == (25,)
        for feature in self.feature_config.get_context_features("node_name"):
            assert features[feature].shape == (1,)
        for feature in self.feature_config.get_sequence_features("node_name"):
            if feature != "clicked":
                assert features[feature].shape == (25,)
        assert labels.shape == (25,)

        # Check tensor shapes when pad_sequence is False
        self.parser.pad_sequence = False
        features, labels = self.parser.get_parse_fn()(self.proto)

        assert features["mask"].shape == (2,)
        for feature in self.feature_config.get_context_features("node_name"):
            assert features[feature].shape == (1,)
        for feature in self.feature_config.get_sequence_features("node_name"):
            if feature != "clicked":
                assert features[feature].shape == (2,)
        assert labels.shape == (2,)
        self.pad_sequence = True



from argparse import Namespace

from ml4ir.base.config.keys import TFRecordTypeKey

from ml4ir.applications.classification.config.keys import LossKey, MetricKey
from ml4ir.base.config.parse_args import RelevanceArgParser

from typing import List


class ClassificationArgParser(RelevanceArgParser):
    """
    States default arguments for classification model.
    """

    def set_default_args(self):
        super().set_default_args()
        self.set_defaults(
            tfrecord_type=TFRecordTypeKey.EXAMPLE,
            loss_key=LossKey.CATEGORICAL_CROSS_ENTROPY,
            metrics_keys=[MetricKey.CATEGORICAL_ACCURACY,
                          MetricKey.TOP_5_CATEGORICAL_ACCURACY],
            monitor_metric=MetricKey.CATEGORICAL_ACCURACY,
            monitor_mode="max",
            group_metrics_min_queries=25,
            output_name="category_label",
        )


def get_args(args: List[str]) -> Namespace:
    return ClassificationArgParser().parse_args(args)



from tensorflow import train
import tensorflow as tf
import pandas as pd

from ml4ir.base.config.keys import SequenceExampleTypeKey


def _bytes_feature(values):
    """Returns a bytes_list from a string / byte."""
    values = [value.encode("utf-8") for value in values]
    return train.Feature(bytes_list=train.BytesList(value=values))


def _float_feature(values):
    """Returns a float_list from a float / double."""
    return train.Feature(float_list=train.FloatList(value=values))


def _int64_feature(values):
    """Returns an int64_list from a bool / enum / int / uint."""
    return train.Feature(int64_list=train.Int64List(value=values))


def _get_feature_fn(dtype):
    """Returns appropriate feature function based on datatype"""
    if dtype == tf.string:
        return _bytes_feature
    elif dtype == tf.float32:
        return _float_feature
    elif dtype == tf.int64:
        return _int64_feature
    else:
        raise Exception("Feature dtype {} not supported".format(dtype))


def get_example_proto(row, features):
    """
    Get an Example protobuf from a pandas dataframe row

    Parameters
    ----------
    row : pandas DataFrame row
        pandas dataframe row to be converted to an example proto
    features : dict
        dictionary containing configuration for all features

    Returns
    -------
    `Example` protobuffer object
        Example object loaded from the specified row
    """

    features_dict = dict()
    for feature_info in features:
        feature_name = feature_info["name"]
        feature_fn = _get_feature_fn(feature_info["dtype"])
        # FIXME
        # When applying functions with axis=1, pandas performs upcasting,
        # so if we have a mix of floats/ints, converts everything to float
        # that breaks this part of the code. Example:
        # https://stackoverflow.com/questions/47143631/
        # how-do-i-preserve-datatype-when-using-apply-row-wise-in-pandas-dataframe
        if feature_name not in row:
            raise Exception(
                "Could not find column {} in record: {}".format(feature_name, str(row))
            )
        features_dict[feature_name] = feature_fn(
            [row[feature_name]]
            if not pd.isna(row[feature_name])
            else [feature_info["default_value"]]
        )

    return train.Example(features=train.Features(feature=features_dict))


def get_sequence_example_proto(group, context_features, sequence_features):
    """
    Get a SequenceExample protobuf from a dataframe group

    Parameters
    ----------
    group : pandas dataframe group
    context_features : dict
        dictionary containing the configuration for all the context features
    sequence_features : dict
        dictionary containing the configuration for all the sequence features

    Returns
    -------
    `SequenceExample` object
        SequenceExample object loaded the dataframe group
    """
    sequence_features_dict = dict()
    context_features_dict = dict()
    for feature_info in context_features:
        feature_name = feature_info["name"]
        feature_fn = _get_feature_fn(feature_info["dtype"])
        context_features_dict[feature_name] = feature_fn([group[feature_name].tolist()[0]])

    for feature_info in sequence_features:
        feature_name = feature_info["name"]
        feature_fn = _get_feature_fn(feature_info["dtype"])
        if feature_info["tfrecord_type"] == SequenceExampleTypeKey.SEQUENCE:
            sequence_features_dict[feature_name] = train.FeatureList(
                feature=[feature_fn(group[feature_name].tolist())]
            )

    return train.SequenceExample(
        context=train.Features(feature=context_features_dict),
        feature_lists=train.FeatureLists(feature_list=sequence_features_dict),
    )



from tensorflow.keras import Input

from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.model.architectures import architecture_factory
from ml4ir.base.model.scoring.interaction_model import InteractionModel
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.io.file_io import FileIO
from logging import Logger

from typing import Dict, Optional


class ScorerBase(object):
    """
    Base Scorer class that defines the neural network layers that convert
    the input features into scores

    Defines the feature transformation layer(InteractionModel), dense
    neural network layers combined with activation layers and the loss function

    Notes
    -----
    This is an abstract class. In order to use a Scorer, one must define
    and override the `architecture_op` and the `final_activation_op` functions
    """

    def __init__(
        self,
        model_config: dict,
        feature_config: FeatureConfig,
        interaction_model: InteractionModel,
        loss: RelevanceLossBase,
        file_io: FileIO,
        output_name: str = "score",
        logger: Optional[Logger] = None,
    ):
        """
        Constructor method for creating a ScorerBase object

        Parameters
        ----------
        model_config : dict
            Dictionary defining the model layer configuration
        feature_config : `FeatureConfig` object
            FeatureConfig object defining the features and their configurations
        interaction_model : `InteractionModel` object
            InteractionModel that defines the feature transformation layers
            on the input model features
        loss : `RelevanceLossBase` object
            Relevance loss object that defines the final activation layer
            and the loss function for the model
        file_io : `FileIO` object
            FileIO object that handles read and write
        output_name : str, optional
            Name of the output that captures the score computed by the model
        logger : Logger, optional
            Logging handler
        """
        self.model_config = model_config
        self.feature_config = feature_config
        self.interaction_model = interaction_model
        self.loss = loss
        self.file_io = file_io
        self.output_name = output_name

    @classmethod
    def from_model_config_file(
        cls,
        model_config_file: str,
        interaction_model: InteractionModel,
        loss: RelevanceLossBase,
        file_io: FileIO,
        output_name: str = "score",
        feature_config: Optional[FeatureConfig] = None,
        logger: Optional[Logger] = None,
    ):
        """
        Get a Scorer object from a YAML model config file

        Parameters
        ----------
        model_config_file : str
            Path to YAML file defining the model layer configuration
        feature_config : `FeatureConfig` object
            FeatureConfig object defining the features and their configurations
        interaction_model : `InteractionModel` object
            InteractionModel that defines the feature transformation layers
            on the input model features
        loss : `RelevanceLossBase` object
            Relevance loss object that defines the final activation layer
            and the loss function for the model
        file_io : `FileIO` object
            FileIO object that handles read and write
        output_name : str, optional
            Name of the output that captures the score computed by the model
        logger: Logger, optional
            Logging handler

        Returns
        -------
        `ScorerBase` object
            ScorerBase object that computes the scores from the input features of the model
        """
        model_config = file_io.read_yaml(model_config_file)

        return cls(
            model_config=model_config,
            feature_config=feature_config,
            interaction_model=interaction_model,
            loss=loss,
            file_io=file_io,
            output_name=output_name,
            logger=logger
        )

    def __call__(self, inputs: Dict[str, Input]):
        """
        Compute score from input features

        Parameters
        --------
        inputs : dict
            Dictionary of input feature tensors

        Returns
        -------
        scores : Tensor object
            Tensor object of the score computed by the model
        train_features : Tensor object
            Dense feature tensor object that is used to compute the model score
        metadata_features : dict of tensor objects
            Dictionary of tensor objects that are not used for training,
            but can be used for computing loss and metrics
        """
        # Apply feature layer and transform inputs
        train_features, metadata_features = self.interaction_model(inputs)

        # Apply architecture op on train_features
        scores = self.architecture_op(train_features, metadata_features)

        # Apply final activation layer
        scores = self.final_activation_op(scores, metadata_features)

        return scores, train_features, metadata_features

    def architecture_op(self, train_features, metadata_features):
        """
        Define and apply the architecture of the model to produce scores from
        transformed features produced by the InteractionModel

        Parameters
        ----------
        train_features : Tensor object
            Dense feature tensor object that is used to compute the model score
        metadata_features : dict of tensor objects
            Dictionary of tensor objects that are not used for training,
            but can be used for computing loss and metrics

        Returns
        -------
        scores : Tensor object
            Tensor object of the score computed by the model
        """
        raise NotImplementedError

    def final_activation_op(self, scores, metadata_features):
        """
        Define and apply the final activation layer to the scores

        Parameters
        ----------
        scores : Tensor object
            Tensor object of the score computed by the model
        metadata_features : dict of tensor objects
            Dictionary of tensor objects that are not used for training,
            but can be used for computing loss and metrics

        Returns
        -------
        scores : Tensor object
            Tensor object produced by applying the final activation function
            to the scores computed by the model
        """
        raise NotImplementedError


class RelevanceScorer(ScorerBase):
    def architecture_op(self, train_features, metadata_features):
        """
        Define and apply the architecture of the model to produce scores from
        transformed features produced by the InteractionModel

        Parameters
        ----------
        train_features : Tensor object
            Dense feature tensor object that is used to compute the model score
        metadata_features : dict of tensor objects
            Dictionary of tensor objects that are not used for training,
            but can be used for computing loss and metrics

        Returns
        -------
        scores : Tensor object
            Tensor object of the score computed by the model
        """
        return architecture_factory.get_architecture(
            model_config=self.model_config,
            feature_config=self.feature_config,
            file_io=self.file_io,
        )(train_features, metadata_features)

    def final_activation_op(self, scores, metadata_features):
        """
        Define and apply the final activation layer to the scores

        Parameters
        ----------
        scores : Tensor object
            Tensor object of the score computed by the model
        metadata_features : dict of tensor objects
            Dictionary of tensor objects that are not used for training,
            but can be used for computing loss and metrics

        Returns
        -------
        scores : Tensor object
            Tensor object produced by applying the final activation function
            to the scores computed by the model
        """
        return self.loss.get_final_activation_op(self.output_name)(
            scores, mask=metadata_features.get("mask")
        )



from typing import Type, List, Union
from tensorflow.keras.metrics import Metric

from ml4ir.base.features.feature_config import FeatureConfig

from typing import Dict


class MetricState:
    OLD = "old"
    NEW = "new"


def get_metrics_impl(
    metrics: List[Union[str, Type[Metric]]],
    feature_config: FeatureConfig,
    metadata_features: Dict,
    **kwargs
) -> List[Union[Metric, str]]:
    """
    Wrapper function to get Metric objects
    to compute validation and test metrics on RelevanceModel

    Parameters
    ----------
    metrics : list
        List of string metric names or custom keras Metric classes
    feature_config : `FeatureConfig` object
        FeatureConfig object that defines the configuration for each
        feature used in the RelevanceModel
    metadata_features : dict
        Dictionary of feature tensors which are not used for training but can be
        used for computing custom metrics

    Returns
    -------
    list
        List of metric names or Metric instances that will be used for
        computing validation and test metrics on RelevanceModel
    """
    metrics_impl: List[Union[Metric, str]] = list()

    for metric in metrics:
        if isinstance(metric, str):
            # If metric is specified as a string, then do nothing
            metrics_impl.append(metric)
        else:
            # If metric is a class of type Metric
            try:
                metrics_impl.extend(
                    [
                        metric(
                            state=MetricState.OLD,
                            feature_config=feature_config,
                            metadata_features=metadata_features,
                            **kwargs
                        ),
                        metric(
                            state=MetricState.NEW,
                            feature_config=feature_config,
                            metadata_features=metadata_features,
                            **kwargs
                        ),
                    ]
                )
            except TypeError:
                metrics_impl.append(metric())

    return metrics_impl



import unittest
import tensorflow as tf
import numpy as np
import os
import random
import gc
from argparse import Namespace

from ml4ir.base.io.local_io import LocalIO
from ml4ir.base.io.logging_utils import setup_logging

from ml4ir.applications.classification.config.parse_args import get_args
from ml4ir.applications.classification.pipeline import ClassificationPipeline
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.model.relevance_model import RelevanceModel

import warnings

warnings.filterwarnings("ignore")


OUTPUT_DIR = "ml4ir/applications/classification/tests/test_output"
ROOT_DATA_DIR = "ml4ir/applications/classification/tests/data"
FEATURE_CONFIG_FNAME = "feature_config.yaml"
MODEL_CONFIG_FNAME = "model_config.yaml"


class ClassificationTestBase(unittest.TestCase):
    """
    Setting default arguments and context for tests .../classification/tests folder.
    """

    @classmethod
    def setUpClass(
        cls,
        output_dir: str = OUTPUT_DIR,
        root_data_dir: str = ROOT_DATA_DIR,
        feature_config_fname: str = FEATURE_CONFIG_FNAME,
        model_config_fname: str = MODEL_CONFIG_FNAME,
    ):
        cls.output_dir = output_dir
        cls.root_data_dir = root_data_dir
        cls.feature_config_fname = feature_config_fname
        cls.model_config_fname = model_config_fname
        cls.file_io = LocalIO()

        # Make temp output directory
        cls.file_io.make_directory(cls.output_dir, clear_dir=True)

        # Fix random seed values for repeatability
        tf.keras.backend.clear_session()
        np.random.seed(123)
        tf.random.set_seed(123)
        random.seed(123)

        # Setup arguments
        cls.args: Namespace = get_args([])
        cls.args.models_dir = output_dir
        cls.args.logs_dir = output_dir

        # Setting small batch size less than testing data size
        cls.args.batch_size = 32

        # Load feature config
        cls.args.feature_config = os.path.join(
            cls.root_data_dir, "configs", cls.feature_config_fname
        )
        cls.feature_config = cls.file_io.read_yaml(cls.args.feature_config)

        # Load model_config
        cls.args.model_config = os.path.join(
            cls.root_data_dir, "configs", cls.model_config_fname
        )
        cls.model_config = cls.file_io.read_yaml(cls.args.model_config)

        # Setup logging
        outfile: str = os.path.join(cls.args.logs_dir, "output_log.csv")

        cls.logger = setup_logging(reset=True,
                                   file_name=outfile,
                                   log_to_file=True)
        cls.run_default_pipeline(data_format="csv")

    @classmethod
    def run_default_pipeline(cls, data_format: str):
        """Train a model with the default set of args"""
        # Fix random seed values for repeatability
        cls.set_seeds()
        args: Namespace = cls.get_overridden_args(data_format)

        cls.classification_pipeline: ClassificationPipeline = ClassificationPipeline(args=args)
        cls.relevance_dataset: RelevanceDataset = cls.classification_pipeline.get_relevance_dataset()
        cls.classification_model: RelevanceModel = cls.classification_pipeline.get_relevance_model()

        cls.train_metrics = cls.classification_model.fit(dataset=cls.relevance_dataset,
                                                         num_epochs=3,
                                                         models_dir=cls.output_dir)

        cls.global_metrics, cls.grouped_metrics, cls.metrics_dict = \
            cls.classification_model.evaluate(test_dataset=cls.relevance_dataset.test,
                                              logs_dir=cls.args.logs_dir,
                                              group_metrics_min_queries=0)
        cls.predictions = cls.classification_model.predict(test_dataset=cls.relevance_dataset.test)

    @classmethod
    def tearDownClass(cls):
        # Delete output directory
        cls.file_io.rm_dir(cls.output_dir)

        # Delete other temp directories
        cls.file_io.rm_dir(os.path.join(cls.root_data_dir, "csv", "tfrecord"))

        # Clear memory
        tf.keras.backend.clear_session()
        gc.collect()

    @classmethod
    def get_overridden_args(cls, data_format: str = "tfrecord"):
        """Overriding test default setup args from parameters."""
        data_dir = os.path.join(cls.root_data_dir, data_format)
        # Fix random seed values for repeatability

        args: Namespace = cls.args
        # Overriding test default setup args from parameters.
        args.data_dir = data_dir
        args.data_format = data_format
        return args

    @staticmethod
    def set_seeds():
        tf.keras.backend.clear_session()
        np.random.seed(123)
        tf.random.set_seed(123)
        random.seed(123)
        return


if __name__ == "__main__":
    unittest.main()



"""
Setup logging utilities.
"""

import logging
from logging import Logger
import os


class LoggingConstants:
    """Logging constant values"""

    LOGGER_NAME = "ranker"
    FNAME = "ranker.log"
    LOG_FORMAT = "%(levelname)s: %(asctime)s.%(msecs)03d \n%(message)s"
    LOG_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


def setup_logging(
    name: str = LoggingConstants.LOGGER_NAME,
    reset: bool = False,
    file_name: str = LoggingConstants.FNAME,
    log_to_file: bool = True,
) -> Logger:

    if not name:
        name = os.path.basename(__file__).split(".")[0]
    logger = logging.getLogger(name)

    if reset:
        map(logger.removeHandler, logger.handlers[:])
        map(logger.removeFilter, logger.filters[:])

    if len(logger.handlers) >= 1:
        return logger

    if log_to_file:
        file_handler = logging.FileHandler(file_name, mode="w")
        file_handler.setFormatter(
            logging.Formatter(
                fmt=LoggingConstants.LOG_FORMAT, datefmt=LoggingConstants.LOG_DATE_FORMAT
            )
        )
        logger.addHandler(file_handler)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    try:
        from colorlog import ColoredFormatter

        color_formatter = ColoredFormatter(
            "%(log_color)s" + LoggingConstants.LOG_FORMAT, datefmt=LoggingConstants.LOG_DATE_FORMAT
        )
    except ImportError:
        color_formatter = logging.Formatter(
            fmt=LoggingConstants.LOG_FORMAT, datefmt=LoggingConstants.LOG_DATE_FORMAT
        )
    stream_handler.setFormatter(color_formatter)
    logger.addHandler(stream_handler)
    logger.setLevel(logging.DEBUG)

    return logger


def set_log_level(debug, logger) -> Logger:
    """
    Sets StreamHandler to debug level.
    """
    if debug > 0:
        for handler in logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setLevel(logging.DEBUG)
    if debug > 1:
        logger.setLevel(1)
    return logger



import pandas as pd
import json
from tensorflow.keras import Input
from logging import Logger
import tensorflow as tf

from ml4ir.base.data.tfrecord_helper import get_sequence_example_proto
from ml4ir.base.config.keys import (
    FeatureTypeKey,
    TFRecordTypeKey,
    SequenceExampleTypeKey,
)

from typing import List, Dict, Optional


class FeatureConfigKey:
    QUERY_KEY = "query_key"
    LABEL = "label"
    FEATURES = "features"
    RANK = "rank"


class FeatureConfig:
    """
    Class that defines the features and their configurations used for
    training, evaluating and serving a RelevanceModel on ml4ir.

    Attributes
    ----------
    features_dict : dict
        Dictionary of features containing the configuration for every feature
        in the model. This dictionary is used to define the FeatureConfig
        object.
    logger : `Logging` object
        Logging handler to log progress messages
    query_key : dict
        Dictionary containing the feature configuration for the unique data point
        ID, query key
    label : dict
        Dictionary containing the feature configuration for the label field
        for training and evaluating the model
    mask : dict
        Dictionary containing the feature configuration for the computed mask
        field which is used to identify padded values
    features : list of dict
        List of dictionaries containing configurations for all the features
        excluding query_key and label
    all_features : list of dict
        List of dictionaries containing configurations for all the features
        including query_key and label
    train_features : list of dict
        List of dictionaries containing configurations for all the features
        which are used for training, identified by `trainable=False`
    metadata_features : list of dict
        List of dictionaries containing configurations for all the features which
        are NOT used for training, identified by `trainable=False`. These can be
        used for computing custom losses and metrics.
    features_to_log : list of dict
        List of dictionaries containing configurations for all the features which
        will be logged when running `model.predict()`, identified using
        `log_at_inference=True`
    group_metrics_keys : list of dict
        List of dictionaries containing configurations for all the features
        which will be used to compute groupwise metrics
    secondary_labels : list of dict
        List of dictionaries containing configurations for all the features
        which will be used as secondary labels to compute secondary metrics.
        The implementation of the secondary metrics and the usage of the secondary
        labels is up to the users of ml4ir

    Notes
    -----
    Abstract class that is overriden by ExampleFeatureConfig and
    SequenceExampleFeatureConfig for the respective TFRecord types
    """

    def __init__(self, features_dict, logger: Optional[Logger] = None):
        """
        Constructor to instantiate a FeatureConfig object

        Parameters
        ----------
        features_dict : dict
            Dictionary containing the feature configuration for each of the
            model features
        logger : `Logging` object, optional
            Logging object handler for logging progress messages
        """
        self.features_dict = features_dict
        self.logger = logger

        self.initialize_features()
        self.extract_features()
        self.log_initialization()

        if len(self.train_features) == 0:
            raise Exception("No trainable features specified in the feature config")

    def initialize_features(self):
        """
        Initialize the feature attributes with empty lists accordingly
        """
        self.all_features: List[Optional[Dict]] = list()
        self.query_key: Optional[Dict] = None
        self.label: Optional[Dict] = None
        self.mask: Optional[Dict] = None
        self.features: List[Optional[Dict]] = list()

        # Features that can be used for training the model
        self.train_features: List[Dict] = list()

        # Features that provide additional information about the query+records
        self.metadata_features: List[Dict] = list()

        # Features to log at inference time
        self.features_to_log: List[Dict] = list()

        # Features to be used as keys for computing group metrics
        # NOTE: Implementation is open-ended
        self.group_metrics_keys: List[Dict] = list()

        # Features to be used as secondary labels for computing secondary metrics
        # NOTE: Implementation is open-ended
        self.secondary_labels: List[Dict] = list()

    @staticmethod
    def get_instance(feature_config_dict: dict, tfrecord_type: str, logger: Logger):
        """
        Factory method to get `FeatureConfig` object from a dictionary of
        feature configurations based on the TFRecord type

        Parameters
        ----------
        feature_config_dict : dict
            Dictionary containing the feature definitions for all the features
            for the model
        tfrecord_type : {"example", "sequence_example"}
            Type of the TFRecord message type used for the ml4ir RelevanceModel
        logger : `Logging` object
            Logging object handler to log status and progress messages

        Returns
        -------
        `FeatureConfig` object
            ExampleFeatureConfig or SequenceExampleFeatureConfig object computed
            from the feature configuration dictionary
        """
        logger.debug(json.dumps(feature_config_dict, indent=4))
        if tfrecord_type == TFRecordTypeKey.EXAMPLE:
            return ExampleFeatureConfig(feature_config_dict, logger=logger)
        else:
            return SequenceExampleFeatureConfig(feature_config_dict, logger=logger)

    def extract_features(self):
        """
        Extract the features from the input feature config dictionary
        and assign to relevant FeatureConfig attributes
        """
        for mandatory_key in [FeatureConfigKey.LABEL, FeatureConfigKey.FEATURES]:
            if mandatory_key not in self.features_dict:
                raise KeyError(
                    "'%s' key not found in the feature_config specified" % mandatory_key
                )
        if FeatureConfigKey.QUERY_KEY not in self.features_dict:
            if self.logger:
                self.logger.warning(
                    "'%s' key not found in the feature_config specified"
                    % FeatureConfigKey.QUERY_KEY
                )
            self.query_key = None
        else:
            self.query_key = self.features_dict.get(FeatureConfigKey.QUERY_KEY)
            self.all_features.append(self.query_key)

        self.label = self.features_dict.get(FeatureConfigKey.LABEL)
        self.all_features.append(self.label)

        self.features = self.features_dict.get(FeatureConfigKey.FEATURES)
        self.all_features.extend(self.features)

        for feature_info in self.all_features:
            if feature_info.get("trainable", True):
                self.train_features.append(feature_info)
            else:
                self.metadata_features.append(feature_info)

            if feature_info.get("log_at_inference", False):
                self.features_to_log.append(feature_info)

            if feature_info.get("is_group_metric_key", False):
                self.group_metrics_keys.append(feature_info)

            if feature_info.get("is_secondary_label", False):
                self.secondary_labels.append(feature_info)

    def log_initialization(self):
        """
        Log initial state of FeatureConfig object after extracting
        all the attributes
        """
        if self.logger:
            self.logger.info("Feature config loaded successfully")
            self.logger.info(
                "Trainable Features : \n{}".format("\n".join(self.get_train_features("name")))
            )
            self.logger.info("Label : {}".format(self.get_label("name")))
            self.logger.info(
                "Metadata Features : \n{}".format("\n".join(self.get_metadata_features("name")))
            )

    def _get_key_or_dict(self, dict_, key: str = None):
        """
        Wrapper method to return config dictionary or fetch a configuration value
        from the Featureconfig.

        Parameters
        ----------
        dict_ : dict
            Dictionary to fetch the value from
        key : str, optional
            Key to select the value for from the dictionary

        Returns
        -------
        str or bool or dict
            Dictionary value if key is passed, otherwise return input dictionary
        """
        if key:
            if key == "node_name":
                return dict_.get("node_name", dict_["name"])
            else:
                return dict_.get(key)
        else:
            return dict_

    def _get_list_of_keys_or_dicts(self, list_of_dicts, key: str = None):
        """
        Wrapper method to return list of config dictionary or fetch a
        configuration value from the Featureconfig.

        Parameters
        ----------
        list_of_dicts : list of dict
            Dictionary to fetch the value from
        key : str, optional
            Key to select the value for from the dictionary

        Returns
        -------
        list of str or bool or dict
            List of dictionary value if key is passed, otherwise return input dictionary
        """
        if key:
            return [self._get_key_or_dict(f, key) for f in list_of_dicts]
        else:
            return list_of_dicts

    def get_query_key(self, key: str = None):
        """
        Getter method for query_key in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str
            Value from the query_key feature configuration to be fetched

        Returns
        -------
        str or int or bool or dict
            Query key value or entire config dictionary based on if the key is passed
        """
        return self._get_key_or_dict(self.query_key, key=key)

    def get_label(self, key: str = None):
        """
        Getter method for label in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str
            Value from the label feature configuration to be fetched

        Returns
        -------
        str or int or bool or dict
            Label value or entire config dictionary based on if the key is passed
        """
        return self._get_key_or_dict(self.label, key=key)

    def get_mask(self, key: str = None):
        """
        Getter method for mask in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str
            Value from the mask feature configuration to be fetched

        Returns
        -------
        str or int or bool or dict
            Label value or entire config dictionary based on if the key is passed
        """
        return self._get_key_or_dict(self.mask, key=key)

    def get_feature_by_node_name(self, name: str):
        """
        Getter method for feature by node name in FeatureConfig object

        Parameters
        ----------
        name : str
            Name of the feature node name to fetch

        Returns
        -------
        dict
            Feature config dictionary for the name of the feature passed
        """
        for feature_info in self.get_all_features():
            if feature_info["node_name"] == name:
                return feature_info

        raise KeyError("No feature with node name {} in FeatureConfig".format(name))

    def get_feature(self, name: str):
        """
        Getter method for feature in FeatureConfig object

        Parameters
        ----------
        name : str
            Name of the feature to fetch

        Returns
        -------
        dict
            Feature config dictionary for the name of the feature passed
        """
        for feature_info in self.get_all_features():
            if feature_info["name"] == name:
                return feature_info

        raise KeyError("No feature named {} in FeatureConfig".format(name))

    def feature_exists(self, name: str, trainable=True):
        """
        Check if a feature exists in FeatureConfig object

        Parameters
        ----------
        name : str
            Name of the feature to fetch

        Returns
        -------
        Boolean
            If a feature exists
        """
        for feature_info in self.get_all_features():
            if feature_info["name"] == name and feature_info["trainable"] == trainable:
                return True

        return False

    def set_feature(self, name: str, new_feature_info: dict):
        """
        Setter method to set the feature_info of a feature in the FeatureConfig
        as specified by the name argument

        Parameters
        ----------
        name : str
            name of feature whose feature_info is to be updated
        new_feature_info : dict
            dictionary used to set the feature_info for the
            feature with specified name
        """
        feature_found = False
        for feature_info in self.features_dict["features"]:
            if feature_info["name"] == name:
                feature_found = True
                self.features_dict["features"].remove(feature_info)
                self.features_dict["features"].append(new_feature_info)

        if feature_found:
            self.initialize_features()
            self.extract_features()
        else:
            raise KeyError("No feature named {} in FeatureConfig".format(name))

    def get_all_features(
        self, key: str = None, include_label: bool = True, include_mask: bool = True
    ):
        """
        Getter method for all_features in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned
        include_label : bool, optional
            Include label in list of features returned
        include_mask : bool, optional
            Include mask in the list of features returned.
            Only applicable with SequenceExampleFeatureConfig currently

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            all features in FeatureConfig
        """
        all_features = self._get_list_of_keys_or_dicts(self.all_features, key=key)

        # Populate features to skip
        features_to_skip = []
        if not include_label:
            features_to_skip.append(self.get_label(key=key))
        if not include_mask:
            features_to_skip.append(self.get_mask(key=key))

        return [f for f in all_features if f not in features_to_skip]

    def get_train_features(self, key: str = None):
        """
        Getter method for train_features in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            trainable features in FeatureConfig
        """
        return self._get_list_of_keys_or_dicts(self.train_features, key=key)

    def get_metadata_features(self, key: str = None):
        """
        Getter method for metadata_features in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            metadata features in FeatureConfig
        """
        return self._get_list_of_keys_or_dicts(self.metadata_features, key=key)

    def get_features_to_log(self, key: str = None):
        """
        Getter method for features_to_log in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            features to be logged at inference
        """
        return self._get_list_of_keys_or_dicts(self.features_to_log, key=key)

    def get_group_metrics_keys(self, key: str = None):
        """
        Getter method for group_metrics_keys in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            features used to compute groupwise metrics
        """
        return self._get_list_of_keys_or_dicts(self.group_metrics_keys, key=key)

    def get_secondary_labels(self, key: str = None):
        """
        Getter method for secondary_labels in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            features to be used as secondary labels
        """
        return self._get_list_of_keys_or_dicts(self.secondary_labels, key=key)

    def get_dtype(self, feature_info: dict):
        """
        Retrieve data type of a feature

        Parameters
        ----------
        feature_info : dict
            Dictionary containing configuration for the feature

        Returns
        -------
        str
            Data type of the feature
        """
        return feature_info["dtype"]

    def get_default_value(self, feature_info):
        """
        Retrieve default value of a feature

        Parameters
        ----------
        feature_info : dict
            Dictionary containing configuration for the feature

        Returns
        -------
        str or int or float
            Default value of the feature
        """
        if feature_info["dtype"] == tf.float32:
            return feature_info["serving_info"].get("default_value", 0.0)
        elif feature_info["dtype"] == tf.int64:
            return feature_info["serving_info"].get("default_value", 0)
        elif feature_info["dtype"] == tf.string:
            return feature_info["serving_info"].get("default_value", "")
        else:
            raise Exception("Unknown dtype {}".format(feature_info["dtype"]))

    def define_inputs(self) -> Dict[str, Input]:
        """
        Define the keras input placeholder tensors for the tensorflow model

        Returns
        -------
        dict
            Dictionary of tensorflow graph input nodes
        """

        def get_shape(feature_info: dict):
            return feature_info.get("shape", (1,))

        inputs: Dict[str, Input] = dict()
        for feature_info in self.get_all_features(include_label=False):
            """
                NOTE: We currently do NOT define label as an input node in the model
                We could do this in the future, to help define more complex loss functions
            """
            node_name = feature_info.get("node_name", feature_info["name"])
            inputs[node_name] = Input(
                shape=get_shape(feature_info), name=node_name, dtype=self.get_dtype(feature_info),
            )

        return inputs

    def create_dummy_protobuf(self, num_records=1, required_only=False):
        """
        Generate a dummy TFRecord protobuffer with dummy values

        Parameters
        ----------
        num_records : int
            Number of records or sequence features per TFRecord message to fetch
        required_only : bool
            Whether to fetch on fields with `required_only=True`

        Returns
        -------
        protobuffer object
            Example or SequenceExample object with dummy values generated
            from the FeatureConfig
        """
        raise NotImplementedError

    def get_hyperparameter_dict(self):
        """
        Create hyperparameter configs to track model metadata for best model selection
        Unwraps the feature config for each of the features to add
        preprocessing_info and feature_layer_info as key value pairs
        that can be tracked across the experiment. This can be used to
        identify the values that were set for the different feature layers
        in a given experiment. Will be used during best model selection and
        Hyper Parameter Optimization.

        Returns
        -------
        dict
            Flattened dictionary of important configuration keys and values
            that can be used for tracking the experiment run
        """
        config = dict()

        config["num_trainable_features"] = len(self.get_train_features())

        for feature_info in self.get_train_features():
            feature_name = feature_info.get("node_name", feature_info["name"])

            # Track preprocessing arguments
            if "preprocessing_info" in feature_info:
                for preprocessing_info in feature_info["preprocessing_info"]:
                    config.update(
                        {
                            "{}_{}_{}".format(feature_name, preprocessing_info["fn"], k): v
                            for k, v in preprocessing_info["args"].items()
                        }
                    )

            # Track feature layer arguments
            if "feature_layer_info" in feature_info:
                feature_layer_info = feature_info["feature_layer_info"]
                if "args" in feature_layer_info:
                    config.update(
                        {
                            "{}_{}".format(feature_name, k): v
                            for k, v in feature_layer_info["args"].items()
                        }
                    )

        return config


class ExampleFeatureConfig(FeatureConfig):
    """
    Class that defines the features and their configurations used for
    training, evaluating and serving a RelevanceModel on ml4ir for
    Example data

    Attributes
    ----------
    features_dict : dict
        Dictionary of features containing the configuration for every feature
        in the model. This dictionary is used to define the FeatureConfig
        object.
    logger : `Logging` object
        Logging handler to log progress messages
    query_key : dict
        Dictionary containing the feature configuration for the unique data point
        ID, query key
    label : dict
        Dictionary containing the feature configuration for the label field
        for training and evaluating the model
    features : list of dict
        List of dictionaries containing configurations for all the features
        excluding query_key and label
    all_features : list of dict
        List of dictionaries containing configurations for all the features
        including query_key and label
    train_features : list of dict
        List of dictionaries containing configurations for all the features
        which are used for training, identified by `trainable=False`
    metadata_features : list of dict
        List of dictionaries containing configurations for all the features which
        are NOT used for training, identified by `trainable=False`. These can be
        used for computing custom losses and metrics.
    features_to_log : list of dict
        List of dictionaries containing configurations for all the features which
        will be logged when running `model.predict()`, identified using
        `log_at_inference=True`
    group_metrics_keys : list of dict
        List of dictionaries containing configurations for all the features
        which will be used to compute groupwise metrics
    secondary_labels : list of dict
        List of dictionaries containing configurations for all the features
        which will be used as secondary labels to compute secondary metrics.
        The implementation of the secondary metrics and the usage of the secondary
        labels is up to the users of ml4ir
    """

    def create_dummy_protobuf(self, num_records=1, required_only=False):
        """Create a SequenceExample protobuffer with dummy values"""
        raise NotImplementedError


class SequenceExampleFeatureConfig(FeatureConfig):
    """
    Class that defines the features and their configurations used for
    training, evaluating and serving a RelevanceModel on ml4ir for
    SequenceExample data

    Attributes
    ----------
    features_dict : dict
        Dictionary of features containing the configuration for every feature
        in the model. This dictionary is used to define the FeatureConfig
        object.
    logger : `Logging` object
        Logging handler to log progress messages
    query_key : dict
        Dictionary containing the feature configuration for the unique data point
        ID, query key
    label : dict
        Dictionary containing the feature configuration for the label field
        for training and evaluating the model
    rank : dict
        Dictionary containing the feature configuration for the rank field
        for training and evaluating the model. `rank` is used to assign an
        ordering to the sequences in the SequenceExample
    mask : dict
        Dictionary containing the feature configuration for the mask field
        for training and evaluating the model. `mask` is used to identify
        which sequence features are padded. A value of 1 represents an
        existing sequence feature and 0 represents a padded sequence feature.
    features : list of dict
        List of dictionaries containing configurations for all the features
        excluding query_key and label
    all_features : list of dict
        List of dictionaries containing configurations for all the features
        including query_key and label
    context_features : list of dict
        List of dictionaries containing configurations for all the features
        which represent the features common to the entire sequence in a
        protobuf message
    sequence_features : list of dict
        List of dictionaries containing configurations for all the features
        which represent the feature unique to a sequence
    train_features : list of dict
        List of dictionaries containing configurations for all the features
        which are used for training, identified by `trainable=False`
    metadata_features : list of dict
        List of dictionaries containing configurations for all the features which
        are NOT used for training, identified by `trainable=False`. These can be
        used for computing custom losses and metrics.
    features_to_log : list of dict
        List of dictionaries containing configurations for all the features which
        will be logged when running `model.predict()`, identified using
        `log_at_inference=True`
    group_metrics_keys : list of dict
        List of dictionaries containing configurations for all the features
        which will be used to compute groupwise metrics
    secondary_labels : list of dict
        List of dictionaries containing configurations for all the features
        which will be used as secondary labels to compute secondary metrics.
        The implementation of the secondary metrics and the usage of the secondary
        labels is up to the users of ml4ir
    """

    def __init__(self, features_dict, logger):
        """
        Constructor to instantiate a FeatureConfig object

        Parameters
        ----------
        features_dict : dict
            Dictionary containing the feature configuration for each of the
            model features
        logger : `Logging` object, optional
            Logging object handler for logging progress messages
        """
        super(SequenceExampleFeatureConfig, self).__init__(features_dict, logger)

    def initialize_features(self):
        """
        Initialize the feature attributes with empty lists accordingly
        """
        super().initialize_features()

        # Feature to capture the rank of the records for a query
        self.rank = None
        # Feature to track padded records
        self.mask = None
        # Features that contain information at the query level common to all records
        self.context_features = list()
        # Features that contain information at the record level
        self.sequence_features = list()

    def extract_features(self):
        """
        Extract the features from the input feature config dictionary
        and assign to relevant FeatureConfig attributes
        """
        super().extract_features()
        try:
            self.rank = self.features_dict.get(FeatureConfigKey.RANK)
            self.all_features.append(self.rank)
        except KeyError:
            self.rank = None
            if self.logger:
                self.logger.warning("'rank' key not found in the feature_config specified")

        for feature_info in self.all_features:
            if feature_info.get("trainable", True):
                self.train_features.append(feature_info)
            else:
                self.metadata_features.append(feature_info)

            if feature_info.get("tfrecord_type") == SequenceExampleTypeKey.CONTEXT:
                self.context_features.append(feature_info)
            elif feature_info.get("tfrecord_type") == SequenceExampleTypeKey.SEQUENCE:
                self.sequence_features.append(feature_info)
            else:
                raise TypeError(
                    "tfrecord_type should be either context or sequence "
                    "but is {} for {}".format(
                        feature_info.get("tfrecord_type"), feature_info.get("name")
                    )
                )

            if feature_info.get("log_at_inference", False):
                self.features_to_log.append(feature_info)

            if feature_info.get("is_group_metric_key", False):
                self.group_metrics_keys.append(feature_info)

            if feature_info.get("is_secondary_label", False):
                self.secondary_labels.append(feature_info)

        self.mask = self.generate_mask()
        self.all_features.append(self.get_mask())

    def get_context_features(self, key: str = None):
        """
        Getter method for context_features in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            context features common to all sequence
        """
        return self._get_list_of_keys_or_dicts(self.context_features, key=key)

    def get_sequence_features(self, key: str = None):
        """
        Getter method for sequence_features in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str, optional
            Name of the configuration key to be fetched.
            If None, then entire dictionary for the feature is returned

        Returns
        -------
        list
            Lift of feature configuration dictionaries or values for
            sequence features unique to each sequence
        """
        return self._get_list_of_keys_or_dicts(self.sequence_features, key=key)

    def log_initialization(self):
        """
        Log initial state of FeatureConfig object after extracting
        all the attributes
        """
        if self.logger:
            self.logger.info("Feature config loaded successfully")
            self.logger.info(
                "Trainable Features : \n{}".format("\n".join(self.get_train_features("name")))
            )
            self.logger.info("Label : {}".format(self.get_label("name")))
            self.logger.info(
                "Metadata Features : \n{}".format("\n".join(self.get_metadata_features("name")))
            )
            self.logger.info(
                "Context Features : \n{}".format("\n".join(self.get_context_features("name")))
            )
            self.logger.info(
                "Sequence Features : \n{}".format("\n".join(self.get_sequence_features("name")))
            )

    def generate_mask(self):
        """
        Add mask information used to flag padded records.
        In order to create a batch of sequence examples from n TFRecords,
        we need to make sure that they all have the same number of sequences.
        To do this, we pad sequence records to a fixed max_sequence_size.
        Now, we do not want to use these additional padded sequence records
        to compute metrics and losses. Hence we maintain a boolean mask to
        tell ml4ir the sequence records that were originally present.

        In this method, we add the feature_info for the above mask feature as it
        is not implicitly present in the data.

        Returns
        -------
        dict
            Dictionary configuration for the mask field that captures
            which sequence have been masked in a SequenceExample message
        """
        return {
            "name": "mask",
            "node_name": "mask",
            "trainable": False,
            "dtype": self.get_rank("dtype"),
            "feature_layer_info": {"type": FeatureTypeKey.NUMERIC, "shape": None},
            "serving_info": {"name": "mask", "required": False},
            "tfrecord_type": SequenceExampleTypeKey.SEQUENCE,
        }

    def get_rank(self, key: str = None):
        """
        Getter method for rank in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str
            Value from the rank feature configuration to be fetched

        Returns
        -------
        str or int or bool or dict
            Rank value or entire config dictionary based on if the key is passed
        """
        return self._get_key_or_dict(self.rank, key=key)

    def get_mask(self, key: str = None):
        """
        Getter method for mask in FeatureConfig object
        Can additionally be used to only fetch a particular value from the dict

        Parameters
        ----------
        key : str
            Value from the mask feature configuration to be fetched

        Returns
        -------
        str or int or bool or dict
            Mask value or entire config dictionary based on if the key is passed
        """
        return self._get_key_or_dict(self.mask, key=key)

    def define_inputs(self) -> Dict[str, Input]:
        """
        Define the keras input placeholder tensors for the tensorflow model

        Returns
        -------
        dict
            Dictionary of tensorflow graph input nodes
        """

        def get_shape(feature_info: dict):
            # Setting size to None for sequence features as the num_records is variable
            if feature_info["tfrecord_type"] == SequenceExampleTypeKey.SEQUENCE:
                return feature_info.get("shape", (None,))
            else:
                return feature_info.get("shape", (1,))

        inputs: Dict[str, Input] = dict()
        for feature_info in self.get_all_features(include_label=False):
            """
                NOTE: We currently do NOT define label as an input node in the model
                We could do this in the future, to help define more complex loss functions
            """
            node_name = feature_info.get("node_name", feature_info["name"])
            inputs[node_name] = Input(
                shape=get_shape(feature_info), name=node_name, dtype=self.get_dtype(feature_info),
            )
        return inputs

    def create_dummy_protobuf(self, num_records=1, required_only=False):
        """
        Generate a dummy TFRecord protobuffer with dummy values

        Parameters
        ----------
        num_records : int
            Number of records or sequence features per TFRecord message to fetch
        required_only : bool
            Whether to fetch on fields with `required_only=True`

        Returns
        -------
        protobuffer object
            Example or SequenceExample object with dummy values generated
            from the FeatureConfig
        """
        context_features = [
            f
            for f in self.get_context_features()
            if ((not required_only) or (f["serving_info"].get("required", False)))
        ]
        sequence_features = [
            f
            for f in self.get_sequence_features()
            if ((not required_only) or (f["serving_info"].get("required", False)))
        ]

        dummy_query = dict()
        for feature_info in self.get_all_features():
            dummy_value = self.get_default_value(feature_info)
            feature_node_name = feature_info.get("node_name", feature_info["name"])
            dummy_query[feature_node_name] = [dummy_value] * num_records

        dummy_query_group = pd.DataFrame(dummy_query).groupby(
            self.get_context_features("node_name")
        )

        return dummy_query_group.apply(
            lambda g: get_sequence_example_proto(
                group=g, context_features=context_features, sequence_features=sequence_features,
            )
        ).values[0]



import sys
import ast
from argparse import Namespace
from typing import Union, List, Type, Optional

from tensorflow.keras.metrics import Metric
from tensorflow.keras.optimizers import Optimizer
from ml4ir.applications.classification.config.parse_args import get_args
from ml4ir.applications.classification.model.losses import categorical_cross_entropy
from ml4ir.applications.classification.model.metrics import metrics_factory
from ml4ir.base.data.relevance_dataset import RelevanceDataset
from ml4ir.base.data.kfold_relevance_dataset import KfoldRelevanceDataset
from ml4ir.base.features.preprocessing import get_one_hot_label_vectorizer
from ml4ir.base.model.losses.loss_base import RelevanceLossBase
from ml4ir.base.model.optimizers.optimizer import get_optimizer
from ml4ir.base.model.relevance_model import RelevanceModel
from ml4ir.base.model.scoring.scoring_model import ScorerBase, RelevanceScorer
from ml4ir.base.model.scoring.interaction_model import InteractionModel, UnivariateInteractionModel
from ml4ir.base.pipeline import RelevancePipeline
from ml4ir.applications.classification.model.classification_model import ClassificationModel


class ClassificationPipeline(RelevancePipeline):
    """Base class that defines a pipeline to train, evaluate and save
    a RelevanceModel for classification using ml4ir"""

    def __init__(self, args: Namespace):
        """
        Constructor to create a RelevancePipeline object to train, evaluate
        and save a model on ml4ir.
        This method sets up data, logs, models directories, file handlers used.
        The method also loads and sets up the FeatureConfig for the model training
        pipeline

        Parameters
        ----------
        args: argparse Namespace
            arguments to be used with the pipeline.
            Typically, passed from command line arguments
        """
        self.loss_key = args.loss_key
        super().__init__(args)

    def get_relevance_model(self, feature_layer_keys_to_fns={}) -> RelevanceModel:
        """
        Creates a RelevanceModel that can be used for training and evaluating

        Parameters
        ----------
        feature_layer_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to tensorflow compatible
            function definitions that can now be used in the InteractionModel
            as a feature function to transform input features

        Returns
        -------
        `RelevanceModel`
            RelevanceModel that can be used for training and evaluating
            a classification model

        Notes
        -----
        Override this method to create custom loss, scorer, model objects
        """

        # Define interaction model
        interaction_model: InteractionModel = UnivariateInteractionModel(
            feature_config=self.feature_config,
            feature_layer_keys_to_fns=feature_layer_keys_to_fns,
            tfrecord_type=self.tfrecord_type,
            file_io=self.file_io,
        )

        # Define loss object from loss key
        loss: RelevanceLossBase = categorical_cross_entropy.get_loss(loss_key=self.loss_key)

        # Define scorer
        scorer: ScorerBase = RelevanceScorer(
            feature_config=self.feature_config,
            model_config=self.model_config,
            interaction_model=interaction_model,
            loss=loss,
            output_name=self.args.output_name,
            logger=self.logger,
            file_io=self.file_io,
        )

        # Define metrics objects from metrics keys
        metrics: List[Union[Type[Metric], str]] = [
            metrics_factory.get_metric(metric_key=metric_key) for metric_key in self.metrics_keys
        ]

        # Define optimizer
        optimizer: Optimizer = get_optimizer(model_config=self.model_config)

        # Combine the above to define a RelevanceModel
        relevance_model: RelevanceModel = ClassificationModel(
            feature_config=self.feature_config,
            scorer=scorer,
            metrics=metrics,
            optimizer=optimizer,
            tfrecord_type=self.tfrecord_type,
            model_file=self.args.model_file,
            initialize_layers_dict=ast.literal_eval(self.args.initialize_layers_dict),
            freeze_layers_list=ast.literal_eval(self.args.freeze_layers_list),
            compile_keras_model=self.args.compile_keras_model,
            output_name=self.args.output_name,
            file_io=self.local_io,
            logger=self.logger,
        )
        return relevance_model

    def get_relevance_dataset(
        self, parse_tfrecord=True, preprocessing_keys_to_fns={}
    ) -> RelevanceDataset:
        """
        Create RelevanceDataset object by loading train, test data as tensorflow datasets
        Defines a preprocessing feature function to one hot vectorize
        classification labels

        Parameters
        ----------
        preprocessing_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to function definitions
            that can now be used for preprocessing while loading the
            TFRecordDataset to create the RelevanceDataset object

        Returns
        -------
        `RelevanceDataset` object
            RelevanceDataset object that can be used for training and evaluating
            the model

        Notes
        -----
        Override this method to create custom dataset objects
        """
        # Adding one_hot_vectorizer needed for classification
        preprocessing_keys_to_fns = {
            "one_hot_vectorize_label": get_one_hot_label_vectorizer(
                self.feature_config.get_label(), self.file_io
            )
        }

        # Prepare Dataset
        relevance_dataset = RelevanceDataset(
            data_dir=self.data_dir_local,
            data_format=self.data_format,
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=parse_tfrecord,
            file_io=self.local_io,
            logger=self.logger,
        )

        return relevance_dataset

    def get_kfold_relevance_dataset(
            self,
            num_folds,
            include_testset_in_kfold,
            read_data_sets=False,
            parse_tfrecord=True,
            preprocessing_keys_to_fns={}
    ) -> KfoldRelevanceDataset:
        """
        Create KfoldRelevanceDataset object by loading train, test data as tensorflow datasets
        Defines a preprocessing feature function to one hot vectorize
        classification labels

        Parameters
        ----------
        num_folds: int
            Number of folds in kfold CV
        include_testset_in_kfold: bool
            Whether to include testset in the folds
        read_data_sets: bool
            Whether to read datasets from disk
        preprocessing_keys_to_fns : dict of (str, function)
            dictionary of function names mapped to function definitions
            that can now be used for preprocessing while loading the
            TFRecordDataset to create the KfoldRelevanceDataset object

        Returns
        -------
        `KfoldRelevanceDataset` object
            KfoldRelevanceDataset object that can be used for training and evaluating
            the model in a kfold cross validation mode.

        Notes
        -----
        Override this method to create custom dataset objects
        """

        # Adding one_hot_vectorizer needed for classification
        preprocessing_keys_to_fns = {
            "one_hot_vectorize_label": get_one_hot_label_vectorizer(
                self.feature_config.get_label(), self.file_io
            )
        }

        # Prepare Dataset
        relevance_dataset = KfoldRelevanceDataset(
            data_dir=self.data_dir_local,
            data_format=self.data_format,
            feature_config=self.feature_config,
            tfrecord_type=self.tfrecord_type,
            max_sequence_size=self.args.max_sequence_size,
            batch_size=self.args.batch_size,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            train_pcent_split=self.args.train_pcent_split,
            val_pcent_split=self.args.val_pcent_split,
            test_pcent_split=self.args.test_pcent_split,
            use_part_files=self.args.use_part_files,
            parse_tfrecord=True,
            file_io=self.local_io,
            logger=self.logger,
            non_zero_features_only=self.non_zero_features_only,
            keep_additional_info=self.keep_additional_info,
            num_folds=num_folds,
            include_testset_in_kfold=include_testset_in_kfold,
            read_data_sets=read_data_sets
        )

        return relevance_dataset

    def create_pipeline_for_kfold(self, args):
        """
        Create a ClassificationPipeline object used in running kfold cross validation.
        """
        return ClassificationPipeline(args=args)

    def run_kfold_analysis(self, base_logs_dir, base_run_id, num_folds, metrics):
        # TODO: Implement the kfold CV analysis for classification
        self.logger.warning("Kfold analysis is not implemented for Classification pipeline")


def main(argv):
    # Define args
    args: Namespace = get_args(argv)

    # Initialize Relevance Pipeline and run in train/inference mode
    rp = ClassificationPipeline(args=args)
    rp.run()


if __name__ == "__main__":
    main(sys.argv[1:])



import tensorflow as tf
from tensorflow import TensorSpec, TensorArray

from ml4ir.base.config.keys import ServingSignatureKey
from ml4ir.base.data.tfrecord_reader import get_parse_fn
from ml4ir.base.features.feature_config import FeatureConfig
from ml4ir.base.io.file_io import FileIO


def define_default_signature(model, feature_config):
    """Default serving signature to take each model feature as input and outputs the scores"""
    raise NotImplementedError


def define_tfrecord_signature(
    model,
    tfrecord_type: str,
    feature_config: FeatureConfig,
    preprocessing_keys_to_fns: dict,
    postprocessing_fn=None,
    required_fields_only: bool = True,
    pad_sequence: bool = False,
    max_sequence_size: int = 0,
):
    """
    Serving signature that wraps around the keras model trained as a RelevanceModel
    with a pre-step to parse TFRecords and apply additional feature preprocessing

    Parameters
    ----------
    model : keras Model
        Keras model object to be saved
    tfrecord_type : {"example", "sequence_example"}
        Type of the TFRecord protobuf that the saved model will be used on at serving time
    feature_config : `FeatureConfig` object
        FeatureConfig object that defines the input features into the model
        and the corresponding feature preprocesing functions to be used
        in the serving signature
    preprocessing_keys_to_fns : dict
        Dictionary mapping function names to tf.functions that should be saved in the preprocessing step of the tfrecord serving signature
    postprocessing_fn: function
        custom tensorflow compatible postprocessing function to be used at serving time.
        Saved as part of the postprocessing layer of the tfrecord serving signature
    required_fields_only: bool
        boolean value defining if only required fields
        need to be added to the tfrecord parsing function at serving time
    pad_sequence: bool, optional
        Value defining if sequences should be padded for SequenceExample proto inputs at serving time.
        Set this to False if you want to not handle padded scores.
    max_sequence_size : int, optional
        Maximum sequence size for SequenceExample protobuf
        The protobuf object will be padded or clipped to this value

    Returns
    -------
    `tf.function`
        Serving signature function that accepts a TFRecord string tensor and returns predictions
    """

    # TFRecord Signature
    # Define a parsing function for tfrecord protos
    inputs = feature_config.get_all_features(key="node_name", include_label=False)

    """
    NOTE:
    Setting pad_sequence=False for tfrecord signature as it is used at inference time
    and we do NOT want to score on padded records for performance reasons

    Limitation: This limits the serving signature to only run inference on a single query
    at a time given the current implementation. This is a tricky issue to fix because
    there is no real way to generate a dense tensor of ranking scores from different queries,
    as they might have varying number of records in each of them.

    Workaround: To infer on multiple queries, run predict() on each of the queries separately.
    """

    tfrecord_parse_fn = get_parse_fn(
        feature_config=feature_config,
        tfrecord_type=tfrecord_type,
        preprocessing_keys_to_fns=preprocessing_keys_to_fns,
        max_sequence_size=max_sequence_size,
        required_fields_only=required_fields_only,
        pad_sequence=pad_sequence,
    )

    dtype_map = dict()
    for feature_info in feature_config.get_all_features(include_label=False):
        feature_node_name = feature_info.get("node_name", feature_info["name"])
        dtype_map[feature_node_name] = feature_config.get_dtype(feature_info)

    # Define a serving signature for tfrecord
    @tf.function(input_signature=[TensorSpec(shape=[None], dtype=tf.string)])
    def _serve_tfrecord(protos):
        input_size = tf.shape(protos)[0]
        features_dict = {
            feature: TensorArray(dtype=dtype_map[feature], size=input_size) for feature in inputs
        }

        # Define loop index
        i = tf.constant(0)

        # Define loop condition
        def loop_condition(i, protos, features_dict):
            return tf.less(i, input_size)

        # Define loop body
        def loop_body(i, protos, features_dict):
            features, labels = tfrecord_parse_fn(protos[i])
            for feature, feature_val in features.items():
                features_dict[feature] = features_dict[feature].write(i, feature_val)

            i += 1

            return i, protos, features_dict

        # Parse all SequenceExample protos to get features
        _, _, features_dict = tf.while_loop(
            cond=loop_condition, body=loop_body, loop_vars=[i, protos, features_dict],
        )

        # Convert TensorArray to tensor
        features_dict = {k: v.stack() for k, v in features_dict.items()}

        # Run the model to get predictions
        predictions = model(inputs=features_dict)

        # Define a post hook
        if postprocessing_fn:
            predictions = postprocessing_fn(predictions, features_dict)

        return predictions

    return _serve_tfrecord


def define_serving_signatures(
    model,
    tfrecord_type: str,
    feature_config: FeatureConfig,
    preprocessing_keys_to_fns: dict,
    postprocessing_fn=None,
    required_fields_only: bool = True,
    pad_sequence: bool = False,
    max_sequence_size: int = 0,
):
    """
    Defines all serving signatures for the SavedModel

    Parameters
    ----------
    model : keras Model
        Keras model object to be saved
    tfrecord_type : {"example", "sequence_example"}
        Type of the TFRecord protobuf that the saved model will be used on at serving time
    feature_config : `FeatureConfig` object
        FeatureConfig object that defines the input features into the model
        and the corresponding feature preprocesing functions to be used
        in the serving signature
    preprocessing_keys_to_fns : dict
        Dictionary mapping function names to tf.functions that should be saved in the preprocessing step of the tfrecord serving signature
    postprocessing_fn: function
        custom tensorflow compatible postprocessing function to be used at serving time.
        Saved as part of the postprocessing layer of the tfrecord serving signature
    required_fields_only: bool
        boolean value defining if only required fields
        need to be added to the tfrecord parsing function at serving time
    pad_sequence: bool, optional
        Value defining if sequences should be padded for SequenceExample proto inputs at serving time.
        Set this to False if you want to not handle padded scores.
    max_sequence_size : int, optional
        Maximum sequence size for SequenceExample protobuf
        The protobuf object will be padded or clipped to this value

    Returns
    -------
    list of `tf.function`
        List of serving signature functions that make predictions at serving time
        with optional pre and post wrapper steps

    Notes
    -----
    Currently only supports TFRecord serving signature
    """
    return {
        ServingSignatureKey.TFRECORD: define_tfrecord_signature(
            model=model,
            tfrecord_type=tfrecord_type,
            feature_config=feature_config,
            preprocessing_keys_to_fns=preprocessing_keys_to_fns,
            postprocessing_fn=postprocessing_fn,
            required_fields_only=required_fields_only,
            pad_sequence=pad_sequence,
            max_sequence_size=max_sequence_size,
        )
    }



from tensorflow.keras import callbacks


class DebuggingCallback(callbacks.Callback):
    def __init__(self, logger, logging_frequency=25):
        super(DebuggingCallback, self).__init__()

        self.logger = logger
        self.epoch = 0
        self.logging_frequency = logging_frequency

    def on_train_batch_end(self, batch, logs=None):
        if batch % self.logging_frequency == 0:
            self.logger.info("[epoch: {} | batch: {}] {}".format(self.epoch, batch, logs))

    def on_predict_batch_end(self, batch, logs=None):
        if batch % self.logging_frequency == 0:
            self.logger.info("[batch: {}] {}".format(batch, logs))

    def on_test_batch_end(self, batch, logs=None):
        if batch % self.logging_frequency == 0:
            self.logger.info("[batch: {}] {}".format(batch, logs))

    def on_epoch_end(self, epoch, logs=None):
        self.logger.info("End of Epoch {}".format(self.epoch))
        self.logger.info(logs)

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch = epoch + 1
        self.logger.info("Starting Epoch : {}".format(self.epoch))
        self.logger.info(logs)
        self.logger.info('lr={}'.format(str(self.model.optimizer._decayed_lr(float).numpy())))

    def on_train_begin(self, logs):
        self.logger.info("Training Model")

    def on_test_begin(self, logs):
        self.logger.info("Evaluating Model")

    def on_predict_begin(self, logs):
        self.logger.info("Predicting scores using model")

    def on_train_end(self, logs):
        self.logger.info("Completed training model")
        self.logger.info(logs)

    def on_test_end(self, logs):
        self.logger.info("Completed evaluating model")
        self.logger.info(logs)

    def on_predict_end(self, logs):
        self.logger.info("Completed Predicting scores using model")
        self.logger.info(logs)
